import '../factory.dart';
import '../graph.dart';

extension AtlasGraphFactoryExtension<V, E> on GraphFactory<V, E> {
  /// Returns a graph from "An Atlas of Graphs" by Ronald C. Read and Robin J.
  /// Wilson, Oxford University Press, 1998. [number] is a number between `0`
  /// and `1252`.
  ///
  /// The graphs are sorted by the number of nodes, the number of vertices,
  /// in increasing order of degree sequence, and in increasing number of
  /// automorphisms.
  Graph<V, E> atlas(int number) {
    final encoded = _atlas[number];
    final builder = newBuilder();
    assert(encoded[0] >= 0, 'Invalid number of vertices');
    for (var i = 0; i < encoded[0]; i++) {
      builder.addVertexIndex(i);
    }
    assert(2 + 2 * encoded[1] == encoded.length, 'Invalid number of edges');
    for (var i = 0, o = 2; i < encoded[1]; i++, o += 2) {
      assert(encoded[o] != encoded[o + 1], 'Invalid edge pair');
      builder.addEdgeIndex(encoded[o], encoded[o + 1]);
    }
    return builder.build();
  }

  /// Returns an iterable over all the graphs in the atlas with the provided
  /// [vertexCount] (from `0` to `7`) and [edgeCount] (from `0` to `21`), see
  /// [atlas] for details.
  Iterable<Graph<V, E>> atlasMatching(
      {int? vertexCount, int? edgeCount}) sync* {
    for (var i = 0; i < _atlas.length; i++) {
      final encoded = _atlas[i];
      if ((vertexCount == null || vertexCount == encoded[0]) &&
          (edgeCount == null || edgeCount == encoded[1])) {
        yield atlas(i);
      }
    }
  }
}

// Data from "An Atlas of Graphs" by Ronald C. Read and Robin J. Wilson,
// Oxford University Press, 1998.
//
// The data was converted from the IGraph package (https://igraph.org/), which
// in turn converted it from the NetworkX package (https://networkx.github.io/).
const _atlas = [
  [0, 0],
  [1, 0],
  [2, 0],
  [2, 1, 0, 1],
  [3, 0],
  [3, 1, 1, 2],
  [3, 2, 0, 1, 0, 2],
  [3, 3, 0, 1, 0, 2, 1, 2],
  [4, 0],
  [4, 1, 3, 2],
  [4, 2, 3, 2, 3, 1],
  [4, 2, 0, 1, 3, 2],
  [4, 3, 3, 2, 1, 2, 3, 1],
  [4, 3, 3, 0, 3, 1, 3, 2],
  [4, 3, 0, 1, 1, 2, 0, 3],
  [4, 4, 3, 2, 1, 2, 3, 1, 3, 0],
  [4, 4, 0, 1, 1, 2, 2, 3, 0, 3],
  [4, 5, 0, 1, 0, 2, 0, 3, 1, 2, 2, 3],
  [4, 6, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2],
  [5, 0],
  [5, 1, 4, 3],
  [5, 2, 1, 2, 0, 1],
  [5, 2, 0, 2, 4, 3],
  [5, 3, 1, 2, 0, 1, 2, 0],
  [5, 3, 4, 3, 3, 2, 3, 1],
  [5, 3, 3, 2, 4, 3, 0, 4],
  [5, 3, 1, 2, 0, 1, 4, 3],
  [5, 4, 4, 3, 1, 2, 3, 1, 3, 2],
  [5, 4, 0, 3, 1, 0, 2, 1, 3, 2],
  [5, 4, 4, 3, 4, 0, 4, 1, 4, 2],
  [5, 4, 4, 0, 3, 1, 4, 3, 3, 2],
  [5, 4, 2, 3, 1, 2, 0, 1, 4, 0],
  [5, 4, 1, 2, 0, 1, 2, 0, 4, 3],
  [5, 5, 0, 3, 2, 0, 3, 2, 1, 0, 2, 1],
  [5, 5, 4, 2, 4, 3, 2, 3, 4, 1, 4, 0],
  [5, 5, 0, 1, 1, 2, 2, 3, 0, 4, 0, 2],
  [5, 5, 4, 0, 1, 2, 4, 3, 3, 2, 3, 1],
  [5, 5, 1, 0, 4, 1, 2, 4, 3, 2, 1, 3],
  [5, 5, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4],
  [5, 6, 1, 0, 4, 1, 4, 0, 0, 3, 1, 3, 3, 4],
  [5, 6, 1, 0, 4, 1, 2, 4, 3, 2, 1, 3, 2, 1],
  [5, 6, 1, 0, 4, 1, 2, 4, 3, 2, 1, 3, 3, 4],
  [5, 6, 0, 1, 4, 3, 2, 3, 4, 2, 4, 0, 4, 1],
  [5, 6, 0, 4, 3, 0, 4, 3, 2, 3, 1, 2, 0, 1],
  [5, 6, 2, 1, 0, 2, 3, 0, 1, 3, 4, 1, 0, 4],
  [5, 7, 4, 0, 1, 2, 4, 3, 3, 2, 3, 1, 4, 1, 2, 4],
  [5, 7, 4, 1, 2, 4, 3, 2, 1, 3, 3, 4, 0, 3, 4, 0],
  [5, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1],
  [5, 7, 2, 1, 0, 2, 3, 0, 1, 3, 4, 1, 0, 4, 2, 4],
  [5, 8, 1, 0, 4, 1, 2, 4, 3, 2, 1, 3, 4, 0, 3, 4, 0, 3],
  [5, 8, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3],
  [5, 9, 0, 1, 3, 4, 0, 3, 0, 4, 1, 2, 1, 3, 1, 4, 2, 3, 2, 4],
  [5, 10, 0, 1, 0, 2, 0, 3, 0, 4, 1, 2, 1, 3, 1, 4, 2, 3, 2, 4, 3, 4],
  [6, 0],
  [6, 1, 5, 4],
  [6, 2, 0, 3, 5, 4],
  [6, 2, 1, 3, 1, 2],
  [6, 3, 1, 3, 2, 1, 3, 2],
  [6, 3, 0, 3, 5, 0, 4, 0],
  [6, 3, 4, 3, 5, 4, 0, 5],
  [6, 3, 4, 3, 5, 1, 5, 2],
  [6, 3, 1, 2, 3, 0, 5, 4],
  [6, 4, 0, 3, 4, 0, 5, 4, 0, 5],
  [6, 4, 3, 0, 5, 3, 4, 5, 0, 4],
  [6, 4, 5, 1, 5, 3, 5, 2, 0, 5],
  [6, 4, 4, 3, 3, 1, 4, 0, 3, 2],
  [6, 4, 0, 2, 1, 3, 2, 1, 5, 3],
  [6, 4, 1, 3, 2, 1, 3, 2, 0, 5],
  [6, 4, 1, 2, 0, 3, 5, 0, 4, 0],
  [6, 4, 4, 5, 1, 2, 0, 5, 3, 4],
  [6, 4, 0, 2, 4, 0, 3, 1, 5, 3],
  [6, 5, 3, 0, 5, 3, 4, 5, 0, 4, 5, 0],
  [6, 5, 5, 3, 3, 1, 3, 2, 4, 3, 4, 5],
  [6, 5, 5, 3, 5, 4, 2, 3, 3, 4, 0, 4],
  [6, 5, 4, 3, 1, 2, 4, 0, 3, 2, 3, 1],
  [6, 5, 1, 4, 3, 4, 4, 0, 2, 1, 3, 2],
  [6, 5, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4],
  [6, 5, 5, 3, 5, 4, 5, 0, 5, 1, 5, 2],
  [6, 5, 1, 4, 5, 1, 1, 0, 2, 1, 2, 3],
  [6, 5, 0, 1, 3, 4, 0, 2, 3, 0, 5, 3],
  [6, 5, 1, 0, 2, 1, 2, 4, 1, 3, 5, 3],
  [6, 5, 4, 3, 0, 5, 4, 0, 3, 2, 3, 1],
  [6, 5, 1, 2, 0, 1, 4, 5, 1, 3, 2, 3],
  [6, 5, 0, 1, 0, 5, 2, 3, 3, 4, 4, 5],
  [6, 5, 4, 3, 5, 1, 5, 2, 0, 3, 4, 0],
  [6, 5, 1, 2, 3, 0, 5, 3, 4, 5, 0, 4],
  [6, 6, 0, 3, 5, 0, 4, 5, 3, 4, 5, 3, 4, 0],
  [6, 6, 1, 4, 2, 4, 4, 0, 2, 3, 3, 1, 3, 4],
  [6, 6, 1, 4, 2, 4, 4, 0, 2, 1, 3, 1, 2, 3],
  [6, 6, 2, 0, 5, 4, 4, 3, 5, 3, 4, 0, 2, 4],
  [6, 6, 3, 2, 4, 3, 0, 4, 1, 0, 2, 1, 0, 3],
  [6, 6, 4, 1, 3, 1, 4, 2, 3, 2, 2, 0, 1, 0],
  [6, 6, 5, 2, 5, 3, 5, 4, 3, 4, 5, 1, 5, 0],
  [6, 6, 4, 3, 4, 2, 4, 0, 1, 4, 3, 0, 5, 3],
  [6, 6, 4, 3, 3, 5, 5, 4, 5, 1, 3, 2, 4, 0],
  [6, 6, 4, 2, 1, 2, 4, 3, 4, 1, 4, 0, 0, 5],
  [6, 6, 1, 2, 3, 1, 0, 3, 2, 0, 4, 0, 5, 0],
  [6, 6, 2, 0, 4, 2, 1, 4, 2, 1, 3, 1, 5, 3],
  [6, 6, 1, 2, 3, 1, 0, 3, 2, 0, 4, 0, 5, 3],
  [6, 6, 5, 3, 2, 5, 2, 0, 4, 2, 4, 3, 3, 1],
  [6, 6, 0, 2, 3, 4, 1, 0, 5, 3, 4, 5, 3, 0],
  [6, 6, 1, 2, 3, 0, 5, 3, 4, 5, 0, 4, 5, 0],
  [6, 6, 4, 3, 1, 2, 4, 0, 3, 2, 3, 1, 5, 0],
  [6, 6, 1, 4, 2, 4, 4, 0, 0, 5, 3, 1, 2, 3],
  [6, 6, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5],
  [6, 6, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5],
  [6, 6, 1, 3, 2, 1, 3, 2, 0, 4, 5, 0, 4, 5],
  [6, 7, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 0, 5],
  [6, 7, 1, 4, 2, 4, 2, 1, 3, 1, 2, 3, 2, 0, 0, 1],
  [6, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1],
  [6, 7, 0, 1, 3, 2, 0, 2, 3, 0, 3, 1, 5, 1, 5, 2],
  [6, 7, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 3, 4],
  [6, 7, 1, 0, 4, 1, 2, 4, 3, 2, 5, 1, 2, 5, 1, 2],
  [6, 7, 0, 4, 2, 0, 1, 2, 3, 1, 5, 3, 3, 0, 2, 3],
  [6, 7, 1, 4, 2, 4, 2, 3, 2, 1, 3, 1, 4, 5, 0, 4],
  [6, 7, 1, 0, 4, 1, 2, 4, 3, 2, 5, 1, 2, 5, 4, 5],
  [6, 7, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4],
  [6, 7, 0, 5, 4, 0, 5, 4, 0, 2, 3, 0, 3, 2, 0, 1],
  [6, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5, 4, 1],
  [6, 7, 0, 1, 4, 0, 1, 4, 0, 2, 3, 0, 3, 2, 3, 5],
  [6, 7, 1, 4, 2, 4, 4, 0, 0, 5, 3, 1, 2, 3, 3, 4],
  [6, 7, 2, 0, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1],
  [6, 7, 1, 5, 0, 1, 4, 0, 3, 4, 2, 3, 1, 2, 0, 3],
  [6, 7, 1, 4, 2, 4, 4, 0, 0, 5, 3, 1, 2, 3, 2, 1],
  [6, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 0, 2, 5, 1],
  [6, 7, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3],
  [6, 7, 5, 0, 3, 5, 2, 3, 0, 2, 1, 3, 4, 1, 3, 4],
  [6, 7, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 2, 3],
  [6, 7, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3],
  [6, 7, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3],
  [6, 7, 1, 2, 0, 1, 2, 0, 3, 0, 4, 3, 5, 4, 3, 5],
  [6, 8, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3],
  [6, 8, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 0, 0, 4],
  [6, 8, 1, 2, 3, 1, 0, 3, 1, 0, 2, 0, 3, 2, 5, 3, 4, 0],
  [6, 8, 0, 1, 2, 4, 0, 2, 5, 2, 3, 1, 3, 2, 2, 1, 4, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 1, 5],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4],
  [6, 8, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 2, 1, 5, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0],
  [6, 8, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 3, 0, 5, 1],
  [6, 8, 2, 0, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1, 5, 3],
  [6, 8, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 0, 5, 5, 4],
  [6, 8, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 5, 1, 5, 3],
  [6, 8, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 3, 4],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 2, 0, 2],
  [6, 8, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2],
  [6, 8, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 1, 4, 0, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0],
  [6, 8, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 2, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 5, 5, 3, 1, 5],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 1],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5, 5, 2, 5, 0],
  [6, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 1, 5, 2],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3],
  [6, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 2],
  [6, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4],
  [6, 9, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 4, 5],
  [6, 9, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 4, 3],
  [6, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 5],
  [6, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0],
  [6, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 0, 4, 1, 0, 4, 1],
  [6, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1],
  [6, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4, 4, 0, 5, 0],
  [6, 9, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5],
  [6, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0],
  [6, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1],
  [6, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 2, 0],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2],
  [6, 9, 2, 3, 0, 2, 3, 0, 4, 3, 1, 4, 5, 1, 4, 5, 1, 0, 5, 2],
  [6, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 5, 2, 4, 1],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 0, 2],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 4, 5],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 0, 5],
  [6, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 1, 0],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5],
  [6, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 2, 4],
  [6, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2],
  [6, 10, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2],
  [6, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1],
  [6, 10, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 5, 3, 5, 4],
  [6, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 0, 2, 1, 3, 5, 1],
  [6, 10, 3, 4, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 2, 4, 5, 1, 3, 2, 0, 3],
  [6, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 5, 3, 2, 5, 1, 0],
  [6, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 0, 2, 1, 5],
  [6, 11, 0, 1, 2, 4, 0, 2, 2, 1, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 0, 3],
  [6, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 0, 4, 5],
  [6, 11, 0, 1, 1, 2, 2, 3, 4, 5, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 3, 0, 2],
  [6, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 5, 3, 2, 5, 1, 0, 5, 1],
  [6, 11, 1, 3, 4, 1, 3, 4, 2, 3, 0, 2, 4, 0, 5, 4, 2, 5, 4, 2, 0, 5, 1, 5],
  [6, 11, 3, 4, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 2, 4, 5, 1, 0, 3, 1, 4, 0, 1],
  [6, 11, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 0, 1, 2, 0, 3, 2],
  [6, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 1, 5, 1, 4, 0, 3],
  [
    6,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    4,
    3
  ],
  [
    6,
    12,
    3,
    2,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    2,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    0,
    4
  ],
  [
    6,
    12,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    4,
    5
  ],
  [
    6,
    12,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    2,
    3
  ],
  [
    6,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    2,
    3,
    1,
    4,
    0,
    2,
    4,
    5,
    1,
    0,
    5,
    4,
    5,
    3,
    4,
    5,
    3
  ],
  [
    6,
    13,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    2,
    3,
    0,
    4
  ],
  [
    6,
    13,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    2,
    3,
    1,
    4,
    0,
    2,
    4,
    5,
    1,
    0,
    5,
    4,
    5,
    3,
    4,
    5,
    3,
    3,
    0
  ],
  [
    6,
    14,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    1,
    3,
    2,
    0,
    4,
    0,
    5,
    3
  ],
  [
    6,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    4,
    5
  ],
  [7, 0],
  [7, 1, 6, 5],
  [7, 2, 2, 3, 1, 2],
  [7, 2, 5, 4, 6, 0],
  [7, 3, 0, 4, 4, 2, 2, 0],
  [7, 3, 0, 1, 0, 6, 0, 5],
  [7, 3, 5, 4, 6, 0, 5, 6],
  [7, 3, 3, 2, 1, 2, 5, 6],
  [7, 3, 3, 1, 5, 6, 0, 4],
  [7, 4, 2, 5, 6, 2, 5, 6, 1, 2],
  [7, 4, 1, 2, 4, 1, 5, 4, 2, 5],
  [7, 4, 1, 0, 5, 1, 1, 2, 4, 1],
  [7, 4, 1, 0, 2, 1, 5, 2, 6, 2],
  [7, 4, 3, 4, 2, 3, 1, 2, 0, 1],
  [7, 4, 4, 2, 0, 4, 2, 0, 5, 6],
  [7, 4, 0, 1, 6, 0, 0, 5, 4, 2],
  [7, 4, 3, 1, 5, 4, 6, 5, 0, 6],
  [7, 4, 0, 4, 3, 0, 2, 5, 6, 2],
  [7, 4, 2, 3, 1, 2, 6, 0, 5, 4],
  [7, 5, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0],
  [7, 5, 2, 5, 6, 2, 5, 6, 4, 2, 3, 2],
  [7, 5, 4, 2, 4, 0, 2, 0, 5, 4, 6, 0],
  [7, 5, 2, 5, 6, 2, 5, 6, 1, 2, 0, 1],
  [7, 5, 4, 1, 0, 4, 3, 0, 1, 3, 2, 1],
  [7, 5, 1, 2, 0, 1, 4, 0, 3, 4, 2, 3],
  [7, 5, 5, 1, 5, 0, 2, 5, 3, 5, 4, 5],
  [7, 5, 1, 5, 6, 1, 1, 0, 2, 1, 3, 2],
  [7, 5, 1, 5, 4, 1, 2, 3, 6, 2, 2, 1],
  [7, 5, 1, 5, 6, 1, 1, 2, 2, 3, 4, 3],
  [7, 5, 2, 1, 3, 2, 4, 3, 5, 4, 3, 6],
  [7, 5, 6, 5, 2, 6, 1, 2, 5, 2, 3, 4],
  [7, 5, 4, 3, 5, 4, 6, 5, 0, 6, 1, 0],
  [7, 5, 0, 4, 3, 0, 2, 5, 6, 2, 5, 6],
  [7, 5, 4, 1, 5, 2, 6, 5, 3, 6, 2, 3],
  [7, 5, 1, 4, 3, 1, 1, 0, 2, 1, 6, 5],
  [7, 5, 0, 4, 3, 0, 1, 0, 2, 1, 6, 5],
  [7, 5, 0, 4, 3, 0, 2, 1, 5, 2, 6, 2],
  [7, 5, 6, 5, 3, 4, 2, 3, 1, 2, 0, 1],
  [7, 5, 2, 3, 1, 2, 6, 0, 5, 6, 5, 4],
  [7, 5, 0, 1, 4, 6, 5, 4, 3, 2, 6, 5],
  [7, 6, 1, 5, 6, 1, 5, 6, 2, 5, 1, 2, 6, 2],
  [7, 6, 1, 4, 3, 1, 2, 3, 4, 2, 1, 0, 2, 1],
  [7, 6, 0, 4, 3, 0, 1, 3, 2, 1, 1, 4, 3, 4],
  [7, 6, 5, 2, 4, 5, 2, 4, 3, 2, 6, 3, 2, 6],
  [7, 6, 1, 2, 4, 1, 5, 4, 2, 5, 0, 1, 4, 0],
  [7, 6, 1, 2, 5, 1, 4, 5, 2, 4, 0, 2, 5, 0],
  [7, 6, 2, 5, 6, 2, 5, 6, 2, 4, 1, 2, 3, 2],
  [7, 6, 1, 4, 3, 1, 2, 3, 1, 2, 2, 5, 6, 2],
  [7, 6, 5, 4, 6, 5, 1, 6, 5, 1, 3, 6, 0, 1],
  [7, 6, 6, 5, 1, 6, 5, 1, 3, 1, 0, 3, 1, 4],
  [7, 6, 0, 4, 3, 0, 2, 3, 4, 2, 2, 5, 6, 2],
  [7, 6, 1, 4, 3, 1, 2, 3, 1, 2, 2, 5, 6, 5],
  [7, 6, 2, 3, 1, 2, 3, 6, 5, 4, 6, 5, 5, 2],
  [7, 6, 2, 5, 6, 2, 5, 6, 1, 4, 3, 1, 2, 1],
  [7, 6, 4, 5, 0, 4, 3, 0, 2, 3, 4, 2, 6, 3],
  [7, 6, 0, 4, 3, 0, 1, 3, 6, 5, 1, 4, 1, 0],
  [7, 6, 1, 4, 3, 1, 2, 3, 5, 2, 6, 5, 2, 6],
  [7, 6, 6, 3, 5, 6, 4, 5, 1, 4, 2, 1, 5, 2],
  [7, 6, 1, 0, 3, 1, 6, 3, 5, 6, 4, 5, 1, 4],
  [7, 6, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5],
  [7, 6, 0, 4, 3, 0, 4, 3, 2, 5, 6, 2, 5, 6],
  [7, 6, 6, 3, 0, 6, 6, 2, 5, 6, 6, 1, 4, 6],
  [7, 6, 2, 4, 5, 2, 2, 3, 6, 2, 1, 2, 1, 0],
  [7, 6, 1, 0, 2, 1, 5, 2, 1, 4, 3, 1, 6, 2],
  [7, 6, 1, 0, 2, 1, 3, 6, 1, 3, 4, 1, 5, 4],
  [7, 6, 1, 0, 2, 1, 5, 2, 6, 5, 1, 4, 3, 1],
  [7, 6, 1, 0, 2, 4, 5, 2, 6, 5, 2, 6, 3, 2],
  [7, 6, 4, 0, 1, 4, 3, 1, 2, 1, 5, 2, 6, 2],
  [7, 6, 6, 5, 1, 2, 0, 1, 2, 0, 3, 2, 0, 4],
  [7, 6, 0, 4, 3, 0, 1, 0, 2, 1, 5, 2, 6, 2],
  [7, 6, 1, 0, 3, 1, 6, 3, 2, 6, 4, 1, 5, 4],
  [7, 6, 2, 5, 6, 2, 4, 2, 1, 4, 3, 1, 0, 3],
  [7, 6, 0, 4, 3, 0, 2, 3, 4, 2, 1, 2, 6, 5],
  [7, 6, 0, 4, 3, 0, 2, 1, 5, 2, 6, 5, 2, 6],
  [7, 6, 3, 4, 1, 0, 2, 1, 5, 2, 6, 5, 2, 6],
  [7, 6, 4, 5, 0, 4, 3, 0, 6, 3, 1, 0, 2, 1],
  [7, 6, 2, 5, 6, 2, 5, 6, 1, 4, 3, 1, 1, 0],
  [7, 6, 4, 5, 3, 4, 2, 3, 1, 2, 0, 1, 6, 0],
  [7, 6, 6, 4, 5, 6, 4, 5, 2, 3, 1, 2, 0, 1],
  [7, 6, 0, 1, 4, 0, 2, 3, 5, 2, 6, 5, 3, 6],
  [7, 6, 1, 2, 0, 1, 4, 0, 3, 4, 2, 3, 6, 5],
  [7, 7, 1, 4, 3, 1, 2, 3, 4, 2, 1, 0, 2, 1, 3, 4],
  [7, 7, 1, 2, 5, 1, 4, 5, 2, 4, 0, 2, 5, 0, 5, 2],
  [7, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1],
  [7, 7, 1, 2, 5, 1, 4, 5, 2, 4, 0, 2, 5, 0, 1, 0],
  [7, 7, 0, 4, 3, 0, 2, 3, 4, 2, 2, 5, 6, 2, 2, 0],
  [7, 7, 1, 4, 3, 1, 2, 3, 4, 2, 1, 0, 2, 1, 2, 6],
  [7, 7, 1, 4, 3, 1, 2, 3, 4, 2, 1, 0, 3, 4, 6, 3],
  [7, 7, 0, 4, 3, 0, 2, 3, 4, 2, 2, 5, 6, 2, 3, 4],
  [7, 7, 0, 4, 3, 0, 1, 3, 3, 6, 1, 4, 1, 0, 5, 4],
  [7, 7, 0, 4, 3, 0, 1, 3, 6, 5, 1, 4, 1, 0, 3, 4],
  [7, 7, 5, 2, 4, 5, 2, 4, 3, 2, 6, 3, 2, 6, 2, 1],
  [7, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 0, 2, 2, 5],
  [7, 7, 5, 2, 4, 5, 2, 4, 3, 2, 6, 3, 2, 6, 3, 1],
  [7, 7, 1, 4, 3, 1, 2, 3, 4, 2, 2, 0, 2, 1, 6, 0],
  [7, 7, 1, 2, 5, 1, 4, 5, 2, 4, 0, 2, 5, 0, 3, 5],
  [7, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 0, 2, 3, 5],
  [7, 7, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 0, 2, 1, 5],
  [7, 7, 3, 2, 4, 3, 3, 5, 2, 4, 5, 2, 6, 1, 6, 4],
  [7, 7, 1, 2, 5, 1, 4, 5, 2, 4, 0, 2, 5, 0, 0, 3],
  [7, 7, 3, 4, 1, 3, 2, 1, 6, 2, 5, 6, 1, 5, 4, 1],
  [7, 7, 0, 1, 4, 0, 1, 4, 2, 1, 3, 2, 5, 3, 4, 5],
  [7, 7, 6, 3, 5, 6, 1, 5, 2, 1, 3, 2, 4, 2, 5, 4],
  [7, 7, 1, 2, 4, 1, 5, 4, 6, 5, 3, 6, 2, 3, 5, 2],
  [7, 7, 4, 1, 3, 4, 1, 3, 2, 1, 6, 2, 5, 6, 2, 5],
  [7, 7, 3, 0, 6, 3, 0, 6, 1, 0, 0, 2, 5, 0, 0, 4],
  [7, 7, 1, 5, 6, 1, 1, 2, 3, 1, 4, 3, 1, 4, 4, 0],
  [7, 7, 5, 0, 6, 5, 0, 6, 5, 2, 1, 5, 6, 3, 4, 6],
  [7, 7, 4, 1, 0, 4, 1, 0, 2, 1, 0, 3, 6, 0, 4, 5],
  [7, 7, 5, 2, 6, 5, 2, 6, 2, 4, 3, 2, 1, 0, 2, 1],
  [7, 7, 4, 1, 0, 4, 3, 0, 1, 3, 2, 1, 1, 5, 6, 1],
  [7, 7, 1, 0, 4, 1, 0, 4, 5, 4, 2, 1, 3, 2, 6, 1],
  [7, 7, 0, 1, 4, 0, 1, 4, 2, 1, 3, 2, 5, 4, 6, 4],
  [7, 7, 2, 3, 5, 2, 6, 5, 3, 6, 1, 2, 4, 5, 0, 5],
  [7, 7, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 2, 1, 6, 5],
  [7, 7, 2, 5, 6, 2, 5, 6, 4, 2, 1, 2, 0, 1, 3, 1],
  [7, 7, 2, 5, 6, 2, 4, 2, 1, 4, 3, 1, 2, 3, 0, 1],
  [7, 7, 6, 2, 5, 6, 2, 5, 1, 2, 0, 1, 4, 1, 3, 1],
  [7, 7, 0, 4, 3, 0, 1, 3, 4, 1, 5, 4, 2, 1, 6, 3],
  [7, 7, 2, 5, 6, 2, 5, 6, 4, 5, 3, 6, 1, 2, 0, 1],
  [7, 7, 2, 5, 6, 2, 1, 4, 1, 2, 0, 1, 4, 0, 0, 3],
  [7, 7, 6, 5, 1, 2, 4, 1, 0, 4, 3, 0, 1, 3, 3, 4],
  [7, 7, 4, 1, 0, 4, 1, 0, 3, 6, 2, 3, 0, 2, 5, 0],
  [7, 7, 4, 1, 0, 4, 3, 0, 1, 3, 2, 1, 5, 2, 6, 1],
  [7, 7, 4, 1, 0, 4, 1, 0, 2, 3, 0, 2, 5, 0, 6, 5],
  [7, 7, 0, 1, 5, 0, 6, 5, 3, 6, 2, 3, 0, 2, 4, 0],
  [7, 7, 1, 0, 4, 1, 2, 4, 3, 2, 4, 3, 0, 4, 6, 5],
  [7, 7, 3, 6, 2, 3, 1, 2, 0, 1, 4, 0, 1, 4, 5, 4],
  [7, 7, 1, 0, 5, 1, 6, 5, 2, 6, 1, 2, 3, 2, 4, 3],
  [7, 7, 2, 3, 1, 2, 0, 1, 4, 0, 5, 4, 6, 5, 4, 1],
  [7, 7, 5, 2, 6, 5, 2, 6, 1, 2, 4, 1, 0, 4, 3, 1],
  [7, 7, 2, 3, 1, 2, 0, 1, 4, 0, 5, 4, 6, 5, 5, 2],
  [7, 7, 1, 4, 0, 1, 2, 0, 3, 2, 5, 3, 0, 5, 6, 3],
  [7, 7, 2, 1, 3, 2, 6, 3, 5, 6, 0, 5, 2, 0, 5, 4],
  [7, 7, 5, 2, 6, 5, 2, 6, 1, 2, 0, 1, 4, 0, 3, 0],
  [7, 7, 4, 1, 0, 4, 3, 0, 1, 3, 2, 1, 5, 2, 6, 2],
  [7, 7, 1, 0, 2, 1, 5, 2, 4, 5, 0, 4, 4, 1, 6, 3],
  [7, 7, 2, 5, 6, 2, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0],
  [7, 7, 6, 5, 0, 4, 3, 0, 1, 3, 4, 1, 2, 4, 3, 2],
  [7, 7, 2, 1, 5, 2, 4, 5, 0, 4, 3, 0, 6, 3, 2, 6],
  [7, 7, 4, 0, 3, 4, 1, 3, 2, 1, 5, 2, 6, 5, 2, 6],
  [7, 7, 6, 5, 2, 6, 1, 2, 4, 1, 0, 4, 3, 0, 1, 3],
  [7, 7, 4, 1, 0, 4, 2, 0, 3, 2, 6, 3, 5, 6, 0, 5],
  [7, 7, 0, 4, 3, 0, 4, 3, 2, 1, 5, 2, 6, 5, 2, 6],
  [7, 7, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 0, 6],
  [7, 7, 1, 0, 4, 1, 0, 4, 5, 2, 6, 5, 3, 6, 2, 3],
  [7, 8, 0, 1, 4, 0, 5, 4, 2, 5, 1, 2, 5, 1, 4, 1, 2, 4],
  [7, 8, 4, 1, 5, 4, 2, 5, 1, 2, 0, 1, 5, 0, 0, 4, 2, 0],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 3, 4, 5, 1, 6, 1],
  [7, 8, 4, 1, 5, 4, 2, 5, 1, 2, 5, 1, 6, 5, 2, 4, 3, 2],
  [7, 8, 1, 3, 0, 1, 4, 0, 2, 4, 1, 2, 4, 1, 5, 4, 1, 5],
  [7, 8, 2, 0, 3, 2, 6, 3, 5, 6, 0, 5, 3, 0, 0, 6, 4, 0],
  [7, 8, 1, 0, 2, 1, 5, 2, 4, 5, 0, 4, 2, 0, 5, 0, 6, 5],
  [7, 8, 1, 0, 2, 1, 3, 2, 1, 3, 4, 3, 2, 4, 5, 2, 3, 5],
  [7, 8, 2, 0, 3, 2, 6, 3, 5, 6, 0, 5, 3, 0, 6, 0, 4, 5],
  [7, 8, 1, 0, 2, 1, 4, 3, 1, 5, 4, 1, 2, 4, 5, 2, 3, 5],
  [7, 8, 3, 5, 2, 1, 4, 3, 1, 5, 4, 1, 2, 4, 5, 2, 4, 6],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 3, 4, 2, 1, 5, 2],
  [7, 8, 3, 5, 2, 1, 4, 3, 1, 5, 4, 1, 2, 4, 5, 2, 0, 3],
  [7, 8, 4, 0, 2, 4, 0, 2, 3, 0, 2, 3, 5, 2, 6, 5, 2, 6],
  [7, 8, 3, 2, 6, 3, 5, 6, 2, 5, 0, 2, 5, 0, 4, 5, 2, 4],
  [7, 8, 0, 5, 4, 0, 2, 4, 5, 2, 1, 5, 4, 1, 3, 4, 5, 3],
  [7, 8, 2, 3, 1, 2, 4, 1, 5, 4, 1, 5, 5, 2, 6, 5, 3, 6],
  [7, 8, 5, 2, 4, 5, 0, 4, 3, 0, 6, 3, 2, 6, 4, 2, 3, 2],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 2, 4, 3, 2, 5, 4, 2, 5],
  [7, 8, 5, 6, 2, 5, 6, 2, 3, 2, 4, 3, 0, 4, 3, 0, 2, 4],
  [7, 8, 1, 0, 5, 0, 3, 2, 1, 3, 5, 2, 6, 1, 6, 2, 6, 5],
  [7, 8, 5, 4, 6, 5, 3, 6, 0, 3, 4, 0, 2, 4, 3, 2, 0, 2],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 2, 1, 5],
  [7, 8, 5, 0, 6, 2, 0, 6, 1, 0, 2, 1, 5, 2, 4, 5, 4, 6],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 2, 1, 1, 5, 6, 1],
  [7, 8, 0, 2, 4, 0, 1, 4, 0, 1, 3, 0, 1, 3, 5, 1, 6, 1],
  [7, 8, 4, 2, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 1, 5, 6, 1],
  [7, 8, 0, 4, 3, 0, 4, 3, 1, 4, 3, 1, 1, 5, 2, 1, 6, 1],
  [7, 8, 2, 1, 0, 2, 3, 0, 5, 3, 2, 5, 3, 2, 4, 3, 6, 5],
  [7, 8, 4, 2, 0, 4, 3, 0, 1, 3, 4, 1, 3, 4, 1, 5, 6, 1],
  [7, 8, 2, 1, 0, 2, 3, 0, 5, 3, 2, 5, 6, 5, 4, 3, 5, 0],
  [7, 8, 1, 0, 2, 1, 3, 2, 1, 3, 4, 2, 3, 4, 4, 5, 6, 4],
  [7, 8, 6, 5, 1, 2, 4, 1, 0, 4, 3, 0, 1, 3, 0, 1, 3, 4],
  [7, 8, 0, 1, 6, 5, 2, 3, 6, 4, 6, 3, 6, 2, 6, 0, 6, 1],
  [7, 8, 6, 4, 1, 2, 2, 3, 6, 5, 4, 5, 6, 2, 6, 0, 6, 1],
  [7, 8, 0, 1, 1, 2, 2, 3, 6, 5, 6, 4, 6, 3, 6, 0, 6, 2],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 6, 1, 5, 1, 2, 5],
  [7, 8, 3, 0, 2, 3, 4, 2, 0, 4, 1, 0, 2, 1, 5, 2, 6, 2],
  [7, 8, 2, 1, 3, 2, 6, 3, 5, 6, 0, 5, 2, 0, 5, 2, 4, 5],
  [7, 8, 1, 0, 2, 1, 3, 2, 4, 3, 5, 2, 1, 5, 6, 1, 2, 6],
  [7, 8, 2, 5, 4, 2, 1, 4, 3, 1, 0, 3, 1, 0, 2, 1, 6, 2],
  [7, 8, 4, 5, 0, 4, 3, 0, 2, 3, 4, 2, 1, 4, 3, 1, 6, 3],
  [7, 8, 0, 1, 4, 0, 1, 4, 2, 1, 4, 2, 5, 4, 1, 5, 6, 3],
  [7, 8, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 1, 6, 5, 0],
  [7, 8, 4, 5, 0, 4, 1, 0, 4, 1, 3, 0, 1, 3, 6, 1, 2, 6],
  [7, 8, 2, 5, 4, 2, 0, 4, 1, 0, 4, 1, 3, 0, 1, 3, 6, 1],
  [7, 8, 1, 6, 2, 1, 0, 2, 1, 0, 4, 1, 3, 4, 2, 3, 4, 5],
  [7, 8, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 1, 6, 5, 3],
  [7, 8, 0, 4, 3, 0, 4, 3, 1, 4, 3, 1, 5, 1, 6, 2, 1, 6],
  [7, 8, 2, 3, 1, 2, 0, 1, 5, 0, 4, 5, 0, 4, 2, 0, 6, 5],
  [7, 8, 4, 5, 0, 4, 3, 0, 1, 3, 4, 1, 2, 4, 3, 2, 6, 2],
  [7, 8, 2, 3, 1, 2, 0, 1, 4, 0, 5, 4, 4, 1, 2, 6, 5, 2],
  [7, 8, 0, 1, 1, 2, 2, 3, 6, 3, 4, 5, 6, 2, 6, 0, 6, 1],
  [7, 8, 4, 1, 0, 4, 3, 0, 1, 3, 0, 1, 2, 1, 5, 2, 6, 2],
  [7, 8, 0, 1, 1, 2, 2, 3, 6, 5, 4, 5, 6, 2, 6, 4, 6, 1],
  [7, 8, 0, 1, 4, 0, 0, 2, 5, 0, 6, 5, 3, 6, 2, 3, 5, 2],
  [7, 8, 0, 4, 3, 0, 2, 3, 4, 2, 1, 4, 3, 1, 2, 5, 6, 2],
  [7, 8, 4, 5, 3, 4, 1, 3, 2, 1, 6, 2, 4, 6, 3, 2, 0, 1],
  [7, 8, 1, 0, 2, 6, 3, 2, 4, 3, 5, 2, 1, 5, 6, 1, 6, 5],
  [7, 8, 2, 3, 1, 2, 0, 1, 4, 0, 5, 4, 6, 5, 5, 2, 4, 1],
  [7, 8, 4, 1, 0, 4, 3, 0, 1, 3, 3, 4, 2, 1, 2, 5, 6, 2],
  [7, 8, 0, 6, 4, 0, 1, 4, 3, 1, 0, 3, 2, 4, 3, 2, 5, 2],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 2, 4, 3, 2, 1, 0, 6, 5],
  [7, 8, 0, 1, 4, 0, 3, 2, 6, 3, 5, 6, 2, 5, 6, 2, 3, 5],
  [7, 8, 5, 2, 6, 5, 2, 6, 4, 2, 0, 4, 3, 0, 2, 3, 1, 2],
  [7, 8, 2, 0, 1, 2, 0, 1, 5, 0, 4, 5, 0, 4, 6, 0, 3, 6],
  [7, 8, 0, 1, 2, 0, 3, 2, 2, 1, 1, 4, 5, 4, 5, 3, 1, 6],
  [7, 8, 1, 6, 2, 1, 0, 2, 1, 0, 4, 1, 3, 4, 2, 3, 5, 6],
  [7, 8, 6, 1, 0, 6, 1, 0, 5, 1, 0, 5, 2, 1, 3, 2, 4, 3],
  [7, 8, 6, 5, 2, 6, 1, 2, 4, 1, 3, 4, 0, 3, 4, 0, 2, 4],
  [7, 8, 1, 6, 0, 1, 5, 0, 1, 5, 3, 0, 4, 3, 2, 4, 0, 2],
  [7, 8, 2, 6, 4, 2, 0, 4, 1, 0, 4, 1, 3, 4, 5, 3, 2, 5],
  [7, 8, 1, 0, 2, 1, 6, 2, 5, 6, 1, 5, 4, 1, 3, 4, 2, 3],
  [7, 8, 6, 1, 4, 3, 1, 0, 5, 1, 3, 2, 2, 1, 4, 6, 5, 4],
  [7, 8, 4, 2, 0, 4, 1, 0, 4, 1, 3, 4, 6, 3, 5, 6, 3, 5],
  [7, 8, 4, 1, 2, 4, 0, 2, 6, 0, 3, 6, 0, 3, 5, 0, 4, 5],
  [7, 8, 5, 6, 4, 5, 0, 4, 3, 0, 2, 3, 4, 2, 1, 4, 3, 1],
  [7, 8, 6, 3, 5, 6, 4, 5, 0, 4, 1, 0, 2, 1, 5, 2, 4, 1],
  [7, 8, 0, 1, 2, 0, 3, 2, 2, 1, 1, 4, 5, 4, 5, 3, 4, 6],
  [7, 8, 4, 0, 3, 4, 2, 3, 6, 2, 5, 6, 1, 5, 4, 1, 2, 1],
  [7, 8, 6, 1, 0, 6, 4, 3, 5, 1, 0, 5, 2, 1, 3, 2, 5, 6],
  [7, 8, 6, 2, 5, 6, 3, 5, 6, 3, 4, 3, 0, 4, 1, 0, 4, 1],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 6, 5, 1],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 6, 4, 2],
  [7, 8, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 6, 0, 5],
  [7, 8, 4, 0, 2, 4, 3, 2, 6, 3, 5, 6, 4, 5, 1, 2, 5, 1],
  [7, 8, 5, 1, 2, 4, 3, 2, 0, 3, 5, 0, 4, 5, 1, 2, 0, 6],
  [7, 8, 5, 6, 2, 5, 4, 2, 0, 4, 3, 0, 2, 3, 1, 4, 3, 1],
  [7, 8, 0, 4, 1, 0, 4, 1, 3, 4, 5, 3, 6, 5, 2, 6, 4, 2],
  [7, 8, 0, 1, 6, 5, 2, 3, 3, 4, 6, 4, 0, 5, 6, 2, 6, 1],
  [7, 8, 1, 2, 0, 1, 4, 0, 5, 4, 2, 5, 3, 2, 6, 3, 5, 6],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 6, 1, 6],
  [7, 8, 6, 2, 5, 6, 2, 5, 1, 2, 4, 1, 0, 4, 3, 0, 1, 3],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 5, 6, 1],
  [7, 8, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 3],
  [7, 8, 0, 4, 1, 0, 3, 2, 1, 4, 2, 5, 5, 3, 6, 4, 6, 3],
  [7, 8, 0, 4, 3, 0, 1, 3, 4, 1, 1, 0, 6, 2, 5, 6, 2, 5],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3],
  [7, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 2],
  [7, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 4, 5],
  [7, 9, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 4, 3],
  [7, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 5],
  [7, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 0, 4, 1, 0, 4, 1],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4, 4, 0, 5, 0],
  [7, 9, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1],
  [7, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 2, 0],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2],
  [7, 9, 2, 3, 0, 2, 3, 0, 4, 3, 1, 4, 5, 1, 4, 5, 1, 0, 5, 2],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 5, 2, 4, 1],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 0, 0, 4, 0, 6],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 0, 0, 4, 2, 6],
  [7, 9, 1, 2, 3, 1, 0, 3, 1, 0, 2, 0, 3, 2, 5, 3, 4, 0, 1, 6],
  [7, 9, 0, 1, 2, 4, 0, 2, 3, 1, 3, 2, 2, 1, 4, 1, 5, 2, 2, 6],
  [7, 9, 0, 1, 2, 4, 0, 2, 5, 2, 3, 1, 3, 2, 2, 1, 4, 1, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 1, 5, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 1, 5, 4, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 2, 1, 5, 1, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0, 1, 6],
  [7, 9, 2, 0, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1, 5, 3, 2, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 3, 0, 5, 1, 0, 6],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 0, 0, 4, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 2, 6],
  [7, 9, 2, 0, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1, 5, 3, 5, 6],
  [7, 9, 1, 2, 3, 1, 0, 3, 1, 0, 2, 0, 3, 2, 4, 0, 6, 5, 6, 3],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0, 0, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 6, 5, 4, 0, 3, 2, 4],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 2, 1, 5, 1, 5, 6],
  [7, 9, 2, 0, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1, 5, 3, 4, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0, 2, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 3, 0, 5, 1, 5, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 5, 4, 3, 1, 3, 2, 3, 0, 5, 1, 2, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 4, 0, 3, 1, 3, 2, 5, 1, 5, 3, 0, 6],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 0, 5, 5, 4, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 5, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 3, 4, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 2, 0, 2, 0, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 3, 4, 3, 6],
  [7, 9, 0, 1, 2, 4, 0, 2, 5, 2, 3, 1, 3, 2, 2, 1, 4, 1, 5, 6],
  [7, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 2, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 6, 5, 6, 1],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 3, 4, 2, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 3, 4, 0, 6],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 6, 5, 3, 6, 1, 6, 0, 1, 1, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 0, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 2, 1, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 5, 6],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 6, 5, 3, 6, 1, 6, 0, 1, 0, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 2, 0, 2, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 2, 0, 2, 4, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 2, 1, 2, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 3, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 2, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 5, 1, 3, 1, 3, 2, 2, 1, 6, 0, 6, 4],
  [7, 9, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 0, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 0, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 3, 6],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 6, 5, 3, 6, 1, 6, 0, 1, 2, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 5, 5, 3, 1, 5, 3, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 1, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 1, 4, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 5, 1, 3, 1, 3, 2, 3, 0, 6, 4, 6, 0],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5, 5, 2, 5, 0, 1, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 2, 1, 6, 0],
  [7, 9, 5, 3, 3, 2, 4, 3, 5, 4, 2, 5, 1, 2, 4, 1, 6, 0, 6, 2],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5, 5, 2, 5, 0, 0, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 5, 5, 3, 1, 5, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 4, 6],
  [7, 9, 1, 3, 2, 1, 0, 2, 5, 0, 6, 5, 3, 6, 1, 6, 0, 1, 5, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 1, 5, 2, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 5, 5, 3, 1, 5, 1, 6],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 2, 1, 3, 6],
  [7, 9, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 0, 5, 5, 4, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 5, 5, 3, 1, 5, 0, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 5, 5, 2, 5, 0, 4, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 1, 0, 6],
  [7, 9, 0, 1, 2, 5, 0, 2, 5, 3, 3, 1, 3, 2, 5, 1, 6, 4, 6, 0],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 1, 5, 2, 0, 6],
  [7, 9, 6, 3, 1, 2, 6, 5, 3, 4, 6, 4, 0, 5, 6, 0, 6, 1, 6, 2],
  [7, 9, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 6, 2, 5, 6, 2, 5],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 3, 4, 6, 5, 6, 0],
  [7, 9, 1, 4, 2, 4, 2, 3, 0, 4, 3, 1, 4, 5, 0, 5, 6, 4, 6, 3],
  [7, 9, 4, 1, 5, 4, 6, 5, 3, 6, 2, 3, 1, 2, 5, 2, 0, 5, 2, 0],
  [7, 9, 4, 1, 3, 1, 2, 3, 4, 0, 5, 0, 5, 2, 5, 4, 6, 4, 5, 6],
  [7, 9, 1, 0, 2, 1, 6, 2, 3, 6, 5, 3, 4, 5, 3, 4, 2, 3, 0, 2],
  [7, 9, 0, 2, 5, 0, 1, 5, 2, 1, 4, 2, 5, 4, 6, 5, 3, 6, 2, 3],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 0, 6, 1, 3, 4, 1],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 4, 5, 1, 6, 1, 0, 6],
  [7, 9, 0, 4, 1, 0, 4, 1, 3, 4, 2, 3, 6, 2, 5, 6, 1, 5, 2, 1],
  [7, 9, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 3, 6, 5, 3, 6],
  [7, 9, 6, 5, 3, 6, 2, 3, 0, 4, 0, 5, 1, 0, 2, 0, 1, 2, 5, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 3, 5, 1, 6, 1, 0, 6],
  [7, 9, 5, 2, 6, 5, 3, 6, 2, 3, 0, 2, 4, 0, 5, 4, 1, 5, 0, 1],
  [7, 9, 2, 4, 1, 2, 4, 1, 5, 4, 0, 5, 1, 0, 6, 4, 3, 6, 2, 3],
  [7, 9, 6, 2, 5, 6, 2, 5, 1, 2, 0, 1, 4, 0, 1, 4, 3, 1, 0, 3],
  [7, 9, 0, 5, 6, 0, 1, 6, 4, 1, 2, 4, 3, 2, 1, 3, 5, 1, 6, 5],
  [7, 9, 6, 5, 3, 6, 2, 3, 5, 2, 0, 5, 1, 0, 4, 1, 0, 4, 2, 0],
  [7, 9, 0, 4, 3, 0, 1, 3, 4, 1, 2, 4, 6, 2, 5, 6, 2, 5, 3, 2],
  [7, 9, 1, 0, 4, 1, 5, 4, 0, 5, 6, 0, 3, 6, 2, 3, 0, 2, 3, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 3],
  [7, 9, 0, 1, 1, 2, 2, 3, 0, 3, 4, 1, 5, 4, 5, 3, 6, 0, 6, 4],
  [7, 9, 0, 1, 4, 0, 1, 4, 2, 1, 5, 2, 4, 5, 6, 5, 3, 6, 2, 3],
  [7, 9, 1, 0, 6, 3, 0, 4, 5, 0, 3, 5, 5, 6, 1, 2, 1, 4, 6, 2],
  [7, 9, 6, 2, 5, 6, 2, 5, 1, 2, 0, 3, 4, 0, 1, 4, 3, 1, 3, 4],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 1, 5, 6, 6, 0],
  [7, 9, 0, 4, 1, 0, 2, 1, 3, 2, 0, 3, 2, 4, 5, 4, 6, 5, 3, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 2, 6, 1, 5, 6],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 3, 5, 4, 6, 1, 6, 5],
  [7, 9, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 4, 6, 2],
  [7, 9, 0, 4, 3, 0, 4, 3, 6, 1, 5, 6, 1, 5, 2, 1, 5, 2, 6, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 0, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 4],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 0],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 1, 0],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 2, 4],
  [7, 10, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1],
  [7, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2],
  [7, 10, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 5, 3, 5, 4],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 0, 2, 1, 3, 5, 1],
  [7, 10, 0, 1, 1, 2, 3, 4, 0, 2, 3, 0, 2, 4, 5, 2, 1, 5, 4, 1, 3, 5],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 5, 3, 2, 5, 1, 0],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 2, 2, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 2, 1, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4, 1, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4, 0, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4, 3, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 4, 5, 4, 6],
  [7, 10, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 4, 3, 3, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 5, 2, 6],
  [7, 10, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 4, 3, 2, 6],
  [7, 10, 2, 3, 1, 2, 4, 1, 5, 4, 2, 5, 0, 2, 4, 0, 0, 1, 5, 0, 6, 5],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 0, 4, 5, 6],
  [7, 10, 2, 0, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 4, 3, 4, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 4, 5, 6, 5],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 6, 5],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0, 0, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0, 2, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 1, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0, 3, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 0, 4, 1, 0, 4, 1, 0, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1, 1, 6],
  [7, 10, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4, 4, 0, 5, 0, 0, 6],
  [7, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 3, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0, 5, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 3, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1, 0, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 2, 0, 5, 3, 2, 3, 6, 2],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 6, 4, 6, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 2, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1, 5, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 3, 0, 4, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 3, 6],
  [7, 10, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4, 4, 0, 5, 0, 2, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 2, 0, 5, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 2, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 2, 0, 5, 3, 2, 3, 0, 6],
  [7, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 1, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 4, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 6, 4, 6, 0],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 5, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 0, 4, 1, 0, 4, 1, 5, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 2, 0, 0, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 1, 0, 5, 1, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 5, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1, 0, 6],
  [7, 10, 0, 1, 1, 2, 0, 2, 3, 0, 3, 1, 3, 2, 5, 4, 4, 0, 5, 0, 5, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 6, 5, 6, 4],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 1, 6],
  [7, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 4, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 4, 6],
  [7, 10, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 0, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2, 5, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1, 1, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 3, 6],
  [7, 10, 4, 3, 4, 1, 1, 2, 5, 4, 2, 5, 3, 1, 5, 3, 3, 2, 6, 0, 6, 2],
  [7, 10, 1, 0, 2, 1, 3, 2, 4, 3, 5, 4, 1, 5, 6, 1, 4, 6, 2, 6, 5, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2, 0, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1, 3, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1, 2, 6],
  [7, 10, 0, 1, 2, 5, 0, 2, 3, 0, 3, 1, 3, 2, 2, 1, 5, 1, 6, 5, 6, 4],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 1, 6],
  [7, 10, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 6],
  [7, 10, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 2, 0, 1, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 1, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 4, 2, 5, 2, 1, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 0, 5, 2, 5, 0, 5, 1, 4, 6],
  [7, 10, 2, 3, 0, 2, 3, 0, 4, 3, 1, 4, 5, 1, 4, 5, 1, 0, 5, 2, 4, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 3, 5, 2, 4, 1, 0, 6],
  [7, 10, 4, 0, 1, 4, 3, 1, 2, 3, 1, 2, 6, 1, 0, 6, 5, 0, 1, 5, 0, 1],
  [7, 10, 3, 2, 6, 3, 5, 6, 0, 5, 2, 0, 5, 2, 1, 5, 2, 1, 4, 2, 5, 4],
  [7, 10, 2, 0, 1, 2, 3, 1, 0, 3, 6, 0, 1, 6, 5, 1, 0, 5, 4, 0, 1, 4],
  [7, 10, 6, 4, 1, 2, 6, 5, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 2, 6, 3],
  [7, 10, 0, 1, 6, 5, 2, 3, 3, 4, 6, 4, 0, 5, 6, 0, 6, 1, 6, 2, 6, 3],
  [7, 10, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 0, 5, 5, 2, 6, 1, 2, 6],
  [7, 10, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 0, 5, 2, 6, 2, 0, 6],
  [7, 10, 6, 4, 1, 2, 6, 5, 3, 4, 4, 5, 0, 5, 6, 3, 6, 1, 6, 2, 0, 4],
  [7, 10, 1, 0, 2, 1, 0, 2, 3, 2, 4, 3, 2, 4, 5, 2, 4, 5, 6, 4, 1, 6],
  [7, 10, 0, 1, 0, 3, 0, 4, 0, 5, 0, 6, 1, 2, 1, 3, 1, 4, 2, 5, 2, 6],
  [7, 10, 0, 2, 5, 0, 4, 5, 2, 4, 1, 2, 5, 1, 6, 5, 3, 6, 2, 3, 2, 6],
  [7, 10, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 1, 0, 5, 6, 0, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 2, 6, 4, 0, 2, 4, 0],
  [7, 10, 0, 4, 3, 0, 2, 3, 5, 2, 6, 5, 2, 6, 4, 2, 1, 4, 3, 1, 4, 3],
  [7, 10, 1, 6, 2, 1, 0, 2, 1, 0, 4, 1, 3, 4, 2, 3, 5, 6, 4, 5, 3, 1],
  [7, 10, 6, 5, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 2, 6, 4],
  [7, 10, 0, 1, 6, 5, 2, 3, 3, 4, 6, 4, 0, 5, 6, 0, 6, 1, 6, 2, 5, 3],
  [7, 10, 0, 1, 1, 2, 2, 3, 5, 4, 0, 4, 5, 0, 5, 3, 5, 2, 6, 5, 6, 1],
  [7, 10, 0, 3, 2, 0, 1, 2, 3, 1, 4, 3, 2, 4, 0, 4, 6, 0, 5, 6, 0, 5],
  [7, 10, 0, 3, 2, 0, 1, 2, 3, 1, 4, 3, 0, 5, 0, 4, 6, 0, 5, 6, 1, 4],
  [7, 10, 0, 1, 6, 5, 2, 3, 3, 4, 6, 4, 0, 5, 6, 0, 6, 1, 6, 2, 4, 2],
  [7, 10, 1, 2, 5, 1, 6, 5, 2, 6, 1, 6, 5, 2, 4, 1, 0, 4, 3, 0, 1, 3],
  [7, 10, 4, 2, 6, 2, 5, 3, 4, 1, 2, 0, 6, 3, 5, 2, 0, 1, 0, 4, 6, 0],
  [7, 10, 4, 2, 3, 6, 5, 3, 5, 1, 2, 0, 6, 0, 5, 2, 1, 4, 0, 4, 5, 4],
  [7, 10, 4, 0, 5, 4, 4, 1, 2, 1, 3, 2, 0, 3, 3, 4, 5, 3, 6, 1, 6, 5],
  [7, 10, 0, 4, 1, 0, 2, 1, 4, 2, 3, 4, 5, 3, 4, 5, 5, 2, 6, 3, 2, 6],
  [7, 10, 1, 6, 2, 1, 0, 2, 1, 0, 4, 1, 3, 4, 2, 3, 5, 6, 4, 5, 4, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 3, 6, 1, 6, 5, 5, 1],
  [7, 10, 1, 0, 4, 1, 0, 4, 2, 0, 3, 2, 6, 3, 5, 6, 0, 5, 5, 2, 6, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 3, 5, 1, 5, 4, 6, 1, 5, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 4, 1, 6, 1, 6, 5],
  [7, 10, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 3, 2, 5, 6, 1, 4, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 2, 5, 4, 6, 5, 6, 0, 0, 2],
  [7, 10, 2, 0, 5, 2, 1, 5, 0, 1, 3, 0, 5, 3, 6, 5, 4, 6, 0, 4, 4, 3],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 2, 1, 5, 6, 2, 6, 5],
  [7, 10, 5, 0, 6, 5, 2, 6, 3, 2, 0, 3, 4, 0, 2, 4, 4, 3, 1, 4, 3, 1],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 6, 3, 5, 6, 3, 5, 4, 5, 4, 6],
  [7, 10, 5, 2, 2, 1, 3, 2, 4, 3, 1, 4, 5, 0, 6, 1, 6, 0, 1, 5, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 2, 6, 4, 4, 2, 5, 1],
  [7, 10, 4, 2, 2, 3, 4, 1, 0, 1, 3, 0, 6, 4, 0, 6, 5, 0, 4, 5, 1, 5],
  [7, 10, 2, 1, 5, 2, 3, 5, 0, 3, 4, 0, 6, 4, 3, 6, 1, 3, 4, 1, 5, 4],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 6, 5, 6, 3],
  [7, 10, 0, 1, 4, 0, 1, 4, 2, 1, 5, 2, 4, 5, 6, 5, 3, 6, 2, 3, 5, 3],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 5, 6, 1, 6, 2, 4, 2],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 3, 5, 2, 6, 5, 6, 4],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 3, 6, 2, 4, 0],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 4, 6, 5, 6, 3],
  [7, 10, 0, 5, 6, 0, 1, 6, 0, 1, 1, 5, 2, 1, 3, 2, 4, 3, 6, 4, 4, 5],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 5, 6, 4, 2, 0],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 5, 6, 3],
  [7, 10, 2, 1, 2, 0, 3, 2, 4, 3, 1, 4, 5, 1, 5, 0, 6, 0, 1, 6, 6, 5],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 3, 5, 1, 6, 1, 6, 4, 6, 5],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 2, 4, 1, 6, 0, 5, 6],
  [7, 10, 3, 1, 0, 3, 5, 0, 1, 5, 2, 1, 6, 2, 0, 6, 0, 4, 4, 2, 4, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 4, 6, 1, 6, 2, 6, 5],
  [7, 10, 0, 3, 2, 0, 1, 2, 3, 1, 4, 3, 2, 4, 5, 4, 5, 0, 6, 4, 6, 1],
  [7, 10, 0, 1, 6, 5, 2, 3, 3, 4, 6, 4, 0, 5, 6, 2, 6, 1, 4, 2, 5, 1],
  [7, 10, 5, 2, 6, 5, 2, 6, 4, 2, 0, 4, 3, 0, 2, 3, 1, 0, 1, 3, 4, 1],
  [7, 10, 3, 4, 1, 3, 4, 1, 0, 4, 3, 0, 1, 0, 2, 1, 5, 2, 6, 5, 2, 6],
  [7, 10, 5, 6, 2, 5, 6, 2, 3, 6, 0, 3, 4, 0, 5, 4, 1, 4, 3, 1, 2, 1],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 5, 4, 2],
  [7, 10, 1, 0, 2, 1, 3, 2, 0, 3, 5, 0, 1, 5, 4, 5, 6, 4, 3, 6, 2, 6],
  [7, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 1, 2, 5, 6, 0, 3, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 0, 2, 1, 5],
  [7, 11, 0, 1, 2, 4, 0, 2, 2, 1, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 0, 3],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 0, 4, 5],
  [7, 11, 0, 1, 1, 2, 2, 3, 4, 5, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 3, 0, 2],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 5, 3, 2, 5, 1, 0, 4, 1, 5, 1],
  [7, 11, 1, 4, 1, 5, 1, 6, 2, 3, 2, 5, 2, 6, 3, 4, 3, 6, 4, 5, 4, 6, 5, 6],
  [7, 11, 3, 6, 1, 3, 2, 1, 0, 2, 5, 0, 6, 5, 2, 6, 5, 1, 0, 3, 1, 6, 0, 1],
  [7, 11, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 0, 1, 2, 0, 3, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 5, 2, 1, 5, 1, 4, 0, 3],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 4, 6, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 3, 6, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 2, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 2, 5, 6, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 2, 0, 6],
  [7, 11, 0, 1, 0, 2, 0, 3, 0, 4, 1, 2, 1, 3, 1, 4, 2, 3, 2, 4, 3, 4, 6, 5],
  [7, 11, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 1, 0, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5, 6, 4],
  [7, 11, 1, 5, 4, 1, 0, 4, 5, 0, 2, 5, 4, 2, 3, 4, 5, 3, 4, 5, 1, 0, 1, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 2, 4, 2, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5, 5, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2, 2, 6],
  [7, 11, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2, 6, 1],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2, 3, 6],
  [7, 11, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2, 6, 4],
  [7, 11, 0, 4, 1, 5, 1, 6, 2, 3, 2, 5, 2, 6, 3, 5, 3, 6, 4, 5, 4, 6, 5, 6],
  [7, 11, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2, 0, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2, 0, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1, 1, 6],
  [7, 11, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 5, 4, 3, 5, 1, 5, 0, 6],
  [7, 11, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 2, 6],
  [7, 11, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2, 3, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1, 2, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2, 5, 6],
  [7, 11, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2, 5, 6],
  [7, 11, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 5, 4, 5, 6],
  [7, 11, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1, 5, 6],
  [7, 11, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 4, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 4, 0, 2, 4, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 1, 3, 4, 1, 2, 4, 0, 3, 6, 5, 6, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 5, 3, 5, 4, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 0, 2, 1, 3, 5, 1, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 5, 3, 5, 4, 1, 6],
  [7, 11, 1, 0, 4, 1, 0, 4, 5, 0, 4, 5, 3, 4, 1, 3, 5, 1, 2, 3, 1, 2, 2, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 3, 2, 0, 3, 2, 4, 5, 2, 1, 6],
  [7, 11, 0, 6, 1, 4, 1, 5, 1, 6, 2, 4, 2, 5, 2, 6, 3, 4, 3, 5, 3, 6, 5, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 5, 3, 2, 5, 1, 0, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 0, 2, 1, 3, 5, 1, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 2, 4, 0, 2, 1, 3, 5, 1, 6, 3],
  [7, 11, 4, 3, 0, 4, 1, 0, 2, 1, 3, 2, 0, 5, 5, 3, 0, 3, 1, 5, 5, 2, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 5, 0, 5, 2, 0, 5, 1, 4, 1, 6, 3],
  [7, 11, 3, 4, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 2, 4, 5, 1, 3, 2, 0, 3, 1, 6],
  [7, 11, 1, 3, 2, 1, 0, 2, 5, 0, 4, 5, 3, 4, 4, 1, 5, 3, 2, 5, 1, 0, 6, 2],
  [7, 11, 0, 1, 2, 4, 0, 2, 4, 5, 3, 1, 3, 2, 4, 1, 5, 1, 5, 2, 5, 3, 0, 6],
  [7, 11, 0, 6, 1, 4, 1, 5, 1, 6, 2, 4, 2, 5, 2, 6, 3, 4, 3, 5, 3, 6, 4, 5],
  [7, 11, 6, 5, 0, 6, 5, 0, 1, 5, 6, 1, 2, 6, 5, 2, 3, 5, 6, 3, 4, 6, 5, 4],
  [7, 11, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 1, 2, 5, 6, 2, 1, 6, 3, 1],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 1, 3, 5, 6, 1, 4, 6, 1, 4, 3, 1],
  [7, 11, 1, 4, 2, 3, 4, 2, 0, 6, 4, 5, 6, 5, 3, 1, 6, 4, 3, 0, 3, 6, 4, 3],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 2, 6, 4, 2, 6, 5, 2, 0, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 4, 0, 3, 0, 2, 0, 6, 0, 3, 6],
  [7, 11, 0, 1, 2, 0, 5, 2, 6, 5, 2, 6, 1, 2, 4, 1, 2, 4, 3, 2, 4, 3, 3, 1],
  [7, 11, 4, 5, 1, 4, 2, 1, 3, 2, 6, 3, 5, 6, 2, 5, 4, 2, 3, 5, 0, 5, 2, 0],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 2, 5, 2, 5, 0, 6, 2, 4, 6, 5, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 6, 1, 2, 6, 0, 2, 6, 0, 5, 2, 0, 5],
  [7, 11, 0, 5, 6, 0, 1, 6, 5, 1, 2, 5, 6, 2, 4, 3, 3, 2, 4, 5, 6, 4, 6, 5],
  [7, 11, 0, 5, 6, 0, 1, 6, 5, 1, 2, 5, 6, 2, 3, 6, 5, 3, 4, 5, 6, 4, 4, 3],
  [7, 11, 0, 5, 0, 6, 1, 2, 1, 6, 2, 4, 3, 4, 3, 5, 3, 6, 4, 5, 4, 6, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 0, 2, 2, 4, 5, 2, 4, 5, 6, 5, 6, 0],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 0, 6, 4, 2, 0, 4, 2, 0, 6, 4],
  [7, 11, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 5, 1, 5, 2, 6, 1, 2, 6, 4, 2, 5, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 2, 5, 4, 2, 6, 2, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 4, 6, 2, 0, 2, 4, 0],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 3, 1, 5, 3, 4, 5, 1, 4, 6, 5, 6, 1],
  [7, 11, 0, 4, 3, 0, 4, 3, 2, 4, 3, 2, 1, 3, 2, 1, 4, 1, 5, 2, 6, 5, 2, 6],
  [7, 11, 0, 5, 0, 6, 1, 4, 1, 6, 2, 3, 2, 5, 3, 4, 3, 5, 3, 6, 4, 5, 4, 6],
  [7, 11, 0, 1, 4, 0, 5, 4, 6, 5, 3, 6, 2, 3, 1, 2, 4, 1, 2, 4, 5, 2, 1, 5],
  [7, 11, 0, 4, 3, 0, 4, 3, 2, 4, 6, 2, 1, 6, 5, 1, 2, 5, 3, 2, 1, 3, 4, 1],
  [7, 11, 0, 1, 6, 5, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 6, 1, 6, 2, 6, 3, 6, 4],
  [7, 11, 4, 1, 0, 4, 1, 0, 3, 1, 0, 3, 5, 1, 6, 5, 1, 6, 2, 1, 5, 2, 6, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 5, 4, 6, 5, 4, 6],
  [7, 11, 1, 0, 2, 1, 3, 2, 4, 3, 0, 4, 2, 0, 5, 2, 6, 5, 3, 6, 6, 0, 0, 5],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 0, 2, 4, 0, 6, 4, 0, 6, 3, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 2, 0, 5, 2, 6, 5, 4, 6, 0, 5, 6, 0],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 0, 2, 6, 3, 6, 0, 3, 4, 0],
  [7, 11, 4, 6, 5, 4, 6, 5, 3, 6, 5, 3, 2, 5, 3, 2, 5, 0, 6, 0, 1, 0, 2, 1],
  [7, 11, 2, 0, 4, 2, 5, 4, 3, 5, 1, 3, 0, 1, 2, 1, 3, 2, 6, 3, 5, 6, 4, 3],
  [7, 11, 4, 3, 4, 2, 1, 4, 3, 1, 0, 3, 1, 0, 3, 2, 3, 5, 2, 5, 6, 0, 4, 6],
  [7, 11, 0, 1, 0, 2, 2, 3, 5, 1, 1, 3, 5, 2, 6, 3, 6, 0, 5, 3, 4, 5, 3, 4],
  [7, 11, 4, 0, 1, 4, 6, 1, 0, 6, 3, 0, 1, 3, 5, 1, 0, 5, 6, 5, 2, 3, 0, 2],
  [7, 11, 0, 1, 5, 0, 4, 5, 1, 4, 2, 1, 3, 2, 4, 3, 4, 2, 6, 4, 2, 6, 3, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 3, 5, 6, 3, 4, 6, 5, 6],
  [7, 11, 6, 3, 5, 6, 2, 5, 3, 2, 5, 3, 4, 5, 2, 4, 1, 2, 5, 1, 0, 1, 4, 0],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 1, 4, 2, 4, 3, 5, 1, 6, 5, 4, 6],
  [7, 11, 0, 4, 3, 0, 2, 3, 5, 2, 6, 5, 2, 6, 4, 2, 1, 4, 3, 1, 1, 0, 2, 1],
  [7, 11, 5, 0, 0, 1, 3, 0, 5, 3, 2, 5, 6, 2, 4, 6, 5, 4, 1, 5, 6, 1, 3, 6],
  [7, 11, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 2, 1, 5, 1, 2, 5, 6, 4, 6, 2, 3, 6],
  [7, 11, 3, 2, 6, 3, 5, 6, 0, 5, 2, 0, 1, 2, 0, 1, 5, 1, 2, 5, 4, 2, 6, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 1, 5, 3, 5, 1, 6, 4, 6, 5, 3, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 3, 6, 0, 6, 2, 6, 3, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 6, 5, 4, 6, 3, 6, 1, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 6, 5, 1, 6, 2, 6, 4, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 3, 4, 2, 6, 1, 6, 4, 5, 2, 3, 5],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 6, 1, 4, 6, 6, 5, 2, 6],
  [7, 11, 0, 3, 0, 6, 1, 2, 1, 5, 2, 4, 2, 6, 3, 4, 3, 5, 4, 5, 4, 6, 5, 6],
  [7, 11, 5, 1, 6, 5, 4, 6, 3, 4, 2, 3, 0, 2, 1, 0, 5, 0, 6, 0, 2, 6, 4, 2],
  [7, 11, 1, 0, 2, 1, 3, 2, 4, 3, 0, 4, 5, 2, 3, 5, 6, 0, 6, 5, 6, 2, 3, 6],
  [7, 11, 0, 3, 4, 0, 2, 4, 3, 2, 1, 3, 4, 1, 5, 1, 6, 0, 6, 1, 5, 3, 4, 5],
  [7, 11, 0, 5, 0, 6, 1, 3, 1, 4, 2, 3, 2, 5, 2, 6, 3, 4, 4, 5, 4, 6, 5, 6],
  [7, 11, 0, 2, 1, 0, 2, 1, 0, 3, 3, 1, 5, 4, 5, 3, 6, 4, 6, 2, 6, 0, 1, 6],
  [7, 11, 4, 1, 5, 4, 2, 5, 1, 2, 0, 1, 5, 0, 6, 5, 3, 6, 2, 3, 0, 2, 4, 0],
  [7, 11, 0, 1, 2, 0, 3, 2, 4, 3, 1, 4, 6, 1, 6, 2, 5, 1, 3, 5, 5, 2, 4, 5],
  [7, 11, 0, 5, 0, 6, 1, 4, 1, 6, 2, 4, 2, 5, 2, 6, 3, 4, 3, 5, 3, 6, 4, 5],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 3, 6, 4, 6, 2, 2, 0, 4, 0],
  [7, 11, 0, 2, 1, 0, 2, 1, 0, 3, 3, 1, 5, 4, 5, 3, 6, 4, 6, 2, 4, 0, 1, 4],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 1, 5, 2, 6, 2, 0, 6, 1, 6],
  [7, 11, 0, 3, 4, 0, 1, 4, 3, 1, 4, 3, 0, 1, 2, 4, 6, 2, 5, 6, 2, 5, 1, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 6, 1, 6, 5],
  [7, 11, 4, 1, 5, 4, 2, 5, 1, 2, 0, 1, 4, 0, 5, 0, 6, 5, 3, 6, 2, 3, 5, 3],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 3, 4, 2, 5, 1, 5, 4, 6, 5, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 5, 4, 0, 4, 5, 0, 5, 1, 5, 2, 5, 3, 6, 3, 6, 4],
  [7, 11, 0, 4, 3, 0, 1, 3, 4, 1, 3, 4, 5, 1, 6, 5, 1, 6, 2, 1, 5, 2, 6, 2],
  [7, 11, 4, 1, 0, 4, 3, 0, 2, 5, 5, 4, 6, 5, 2, 6, 1, 2, 3, 1, 6, 3, 2, 3],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 2, 5, 4, 5, 3, 6, 0, 6, 5, 3, 6],
  [7, 11, 5, 2, 2, 4, 5, 3, 4, 1, 5, 4, 0, 1, 3, 0, 0, 2, 6, 2, 6, 3, 0, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 5, 0, 5, 4, 5, 2, 5, 3, 6, 1, 0, 6],
  [7, 11, 0, 3, 0, 4, 1, 2, 1, 5, 1, 6, 2, 4, 2, 6, 3, 5, 3, 6, 4, 5, 5, 6],
  [7, 11, 4, 0, 3, 4, 5, 3, 0, 5, 1, 0, 2, 1, 3, 2, 4, 1, 5, 2, 6, 4, 5, 6],
  [7, 11, 2, 3, 4, 2, 0, 4, 5, 0, 1, 5, 4, 1, 3, 4, 5, 3, 1, 0, 6, 5, 6, 2],
  [7, 11, 4, 1, 0, 4, 3, 0, 4, 3, 5, 4, 6, 5, 2, 6, 1, 2, 3, 1, 6, 3, 2, 5],
  [7, 11, 0, 3, 4, 0, 2, 4, 3, 2, 1, 3, 0, 1, 6, 0, 5, 6, 2, 5, 1, 5, 4, 1],
  [7, 11, 0, 3, 0, 4, 1, 4, 1, 5, 1, 6, 2, 3, 2, 5, 2, 6, 3, 6, 4, 5, 5, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 0, 3, 4, 0, 4, 3, 4, 2, 5, 1, 5, 4, 6, 1, 5, 6],
  [7, 11, 4, 1, 5, 4, 6, 5, 3, 6, 2, 3, 1, 2, 0, 1, 5, 0, 4, 0, 5, 2, 6, 2],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 0, 4, 4, 2, 3, 5, 4, 5, 3, 0, 6, 5, 6, 1],
  [7, 11, 0, 4, 1, 0, 4, 1, 3, 4, 2, 3, 1, 2, 6, 1, 5, 6, 3, 5, 5, 4, 2, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 5, 1, 4, 2, 6, 2, 3, 6, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 0, 5, 6, 2, 1, 6, 5, 2, 4, 1, 0, 3],
  [7, 11, 0, 3, 0, 4, 1, 2, 1, 5, 1, 6, 2, 5, 2, 6, 3, 5, 3, 6, 4, 5, 4, 6],
  [7, 11, 0, 1, 1, 2, 2, 3, 5, 4, 0, 4, 5, 3, 5, 1, 6, 3, 6, 4, 4, 2, 0, 3],
  [7, 11, 0, 4, 0, 5, 0, 6, 1, 3, 1, 5, 1, 6, 2, 3, 2, 4, 2, 6, 3, 6, 4, 5],
  [7, 11, 4, 3, 2, 4, 3, 2, 0, 3, 2, 1, 5, 4, 5, 0, 6, 4, 6, 1, 1, 5, 0, 6],
  [7, 11, 6, 4, 3, 6, 1, 3, 4, 1, 0, 4, 2, 0, 3, 2, 0, 1, 5, 0, 6, 5, 5, 2],
  [7, 11, 6, 1, 2, 6, 1, 2, 0, 1, 3, 0, 4, 3, 5, 4, 3, 5, 2, 0, 4, 0, 5, 6],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    4,
    3,
    0,
    2
  ],
  [
    7,
    12,
    3,
    6,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    6,
    5,
    2,
    6,
    5,
    1,
    0,
    3,
    1,
    6,
    0,
    1,
    0,
    6
  ],
  [
    7,
    12,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    5,
    1,
    2,
    4
  ],
  [
    7,
    12,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    2,
    3
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    2,
    3,
    1,
    4,
    0,
    2,
    4,
    5,
    1,
    0,
    5,
    4,
    5,
    3,
    4,
    5,
    3
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    0,
    2,
    1,
    5,
    6,
    1
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    0,
    2,
    4,
    6,
    5,
    3
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    2,
    1,
    3,
    1,
    3,
    2,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    1,
    6
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    2,
    1,
    3,
    1,
    3,
    2,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    0,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    0,
    4,
    5,
    6,
    1
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    2,
    1,
    3,
    1,
    3,
    2,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    0,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    0,
    4,
    5,
    0,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    0,
    4,
    5,
    2,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    0,
    2,
    1,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    0,
    2,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    0,
    2,
    6,
    1,
    6,
    5
  ],
  [
    7,
    12,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    1,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    1,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    3,
    5,
    0,
    3,
    5,
    6
  ],
  [
    7,
    12,
    3,
    6,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    6,
    5,
    2,
    6,
    5,
    1,
    0,
    3,
    1,
    6,
    0,
    1,
    1,
    4
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    3,
    5,
    0,
    3,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    2,
    1,
    3,
    1,
    3,
    2,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    4,
    6
  ],
  [
    7,
    12,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    1,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    2,
    6
  ],
  [
    7,
    12,
    3,
    6,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    6,
    5,
    2,
    6,
    5,
    1,
    0,
    3,
    1,
    6,
    0,
    1,
    0,
    4
  ],
  [
    7,
    12,
    1,
    3,
    4,
    1,
    3,
    4,
    2,
    3,
    0,
    2,
    4,
    0,
    5,
    4,
    2,
    5,
    4,
    2,
    0,
    5,
    1,
    5,
    3,
    6
  ],
  [
    7,
    12,
    1,
    3,
    4,
    1,
    3,
    4,
    2,
    3,
    0,
    2,
    4,
    0,
    5,
    4,
    2,
    5,
    4,
    2,
    0,
    5,
    1,
    5,
    0,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    0,
    4,
    5,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    0,
    3,
    5,
    6
  ],
  [
    7,
    12,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    4,
    6
  ],
  [
    7,
    12,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    0,
    6
  ],
  [
    7,
    12,
    3,
    6,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    6,
    5,
    2,
    6,
    5,
    1,
    0,
    3,
    1,
    6,
    0,
    1,
    4,
    5
  ],
  [
    7,
    12,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    0,
    3,
    0,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    0,
    2,
    5,
    6
  ],
  [
    7,
    12,
    0,
    3,
    6,
    0,
    5,
    6,
    3,
    5,
    1,
    3,
    6,
    1,
    4,
    6,
    3,
    4,
    2,
    3,
    6,
    2,
    3,
    6,
    5,
    4
  ],
  [
    7,
    12,
    0,
    1,
    4,
    0,
    5,
    4,
    1,
    5,
    4,
    1,
    2,
    4,
    1,
    2,
    6,
    1,
    4,
    6,
    2,
    6,
    3,
    2,
    4,
    3
  ],
  [
    7,
    12,
    4,
    1,
    3,
    2,
    3,
    0,
    4,
    2,
    5,
    1,
    5,
    0,
    3,
    5,
    4,
    3,
    5,
    4,
    6,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    3,
    1,
    3,
    2,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    5,
    6,
    3,
    1,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    3,
    1,
    3,
    2,
    6,
    3,
    0,
    6,
    5,
    0,
    1,
    5,
    4,
    1,
    2,
    4
  ],
  [
    7,
    12,
    6,
    2,
    5,
    6,
    3,
    5,
    2,
    3,
    1,
    2,
    4,
    1,
    5,
    4,
    2,
    5,
    4,
    2,
    0,
    4,
    1,
    0,
    5,
    1
  ],
  [
    7,
    12,
    5,
    4,
    6,
    5,
    3,
    6,
    4,
    3,
    0,
    4,
    3,
    0,
    1,
    3,
    0,
    1,
    4,
    1,
    3,
    5,
    2,
    3,
    4,
    2
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    5,
    0,
    1,
    5,
    6,
    1,
    0,
    6
  ],
  [
    7,
    12,
    1,
    2,
    0,
    1,
    2,
    0,
    3,
    2,
    0,
    3,
    1,
    3,
    4,
    2,
    0,
    4,
    5,
    4,
    2,
    5,
    6,
    2,
    1,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    2,
    4,
    5,
    6,
    4,
    3,
    6
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    5,
    0,
    1,
    5,
    6,
    1,
    2,
    6
  ],
  [
    7,
    12,
    1,
    2,
    0,
    1,
    2,
    0,
    3,
    2,
    0,
    3,
    1,
    3,
    4,
    2,
    0,
    4,
    6,
    4,
    2,
    6,
    5,
    2,
    4,
    5
  ],
  [
    7,
    12,
    1,
    0,
    2,
    1,
    0,
    2,
    3,
    0,
    4,
    3,
    0,
    4,
    5,
    0,
    3,
    5,
    4,
    5,
    6,
    4,
    3,
    6,
    6,
    0
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    3,
    1,
    3,
    2,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    6,
    5,
    2,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    6,
    4,
    0,
    6,
    5,
    0,
    3,
    5
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    1,
    5,
    6,
    1,
    0,
    6,
    5,
    0,
    4,
    5
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    6,
    1,
    0,
    6,
    5,
    0,
    2,
    5
  ],
  [
    7,
    12,
    5,
    4,
    3,
    5,
    4,
    3,
    6,
    4,
    3,
    6,
    2,
    3,
    4,
    2,
    0,
    4,
    3,
    0,
    1,
    0,
    2,
    1,
    6,
    2
  ],
  [
    7,
    12,
    4,
    1,
    3,
    2,
    3,
    0,
    4,
    2,
    5,
    1,
    5,
    0,
    3,
    5,
    4,
    3,
    5,
    4,
    6,
    5,
    0,
    6,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    6,
    1,
    0,
    6,
    5,
    0,
    1,
    5
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    3,
    1,
    5,
    1,
    6,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    5,
    4,
    3,
    5,
    6,
    3,
    4,
    6
  ],
  [
    7,
    12,
    1,
    0,
    2,
    1,
    3,
    2,
    0,
    3,
    4,
    3,
    1,
    4,
    4,
    0,
    5,
    4,
    0,
    5,
    3,
    5,
    6,
    0,
    1,
    6
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    6,
    2,
    4,
    6,
    5,
    0,
    1,
    5
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    5,
    4,
    2,
    5,
    1,
    5,
    6,
    1,
    5,
    6
  ],
  [
    7,
    12,
    0,
    2,
    1,
    0,
    2,
    1,
    0,
    3,
    3,
    1,
    4,
    2,
    5,
    3,
    6,
    1,
    6,
    4,
    4,
    0,
    3,
    4,
    1,
    5
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    3,
    2,
    4,
    3,
    4,
    1,
    0,
    4,
    5,
    2,
    5,
    4,
    6,
    0,
    3,
    6
  ],
  [
    7,
    12,
    5,
    0,
    2,
    5,
    6,
    2,
    3,
    6,
    2,
    3,
    1,
    2,
    0,
    1,
    4,
    0,
    1,
    4,
    5,
    1,
    6,
    5,
    0,
    2
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    6,
    4,
    2,
    6,
    5,
    2,
    3,
    5
  ],
  [
    7,
    12,
    0,
    2,
    1,
    0,
    5,
    1,
    3,
    5,
    6,
    3,
    2,
    6,
    4,
    2,
    3,
    4,
    5,
    4,
    2,
    5,
    1,
    2,
    4,
    1
  ],
  [
    7,
    12,
    0,
    2,
    1,
    0,
    2,
    1,
    0,
    3,
    3,
    1,
    4,
    0,
    1,
    4,
    4,
    2,
    3,
    4,
    5,
    4,
    6,
    5,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    3,
    1,
    3,
    2,
    6,
    1,
    2,
    6,
    4,
    6,
    3,
    4,
    5,
    3,
    6,
    5
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    5,
    0,
    6,
    5,
    0,
    6,
    3,
    4
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    1,
    2,
    5,
    6,
    0,
    3,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    5,
    3,
    5,
    1,
    5,
    4,
    6,
    3,
    4,
    6,
    1,
    6,
    6,
    5
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    5,
    2,
    6,
    5,
    0,
    6,
    6,
    2,
    0,
    5,
    1,
    5,
    6,
    1
  ],
  [
    7,
    12,
    0,
    1,
    6,
    5,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    6,
    0,
    6,
    1,
    6,
    2,
    6,
    3,
    6,
    4,
    5,
    1
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    5,
    0,
    5,
    1,
    5,
    2,
    5,
    3,
    5,
    4,
    6,
    1,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    1,
    4,
    5,
    6,
    4,
    5,
    6
  ],
  [
    7,
    12,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    1,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    3,
    6,
    1
  ],
  [
    7,
    12,
    0,
    4,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    2,
    1,
    4,
    6,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    6,
    1,
    2,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    5,
    4,
    0,
    4,
    5,
    3,
    5,
    1,
    6,
    3,
    6,
    4,
    4,
    2,
    0,
    3,
    4,
    3
  ],
  [
    7,
    12,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    1,
    6,
    5
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    1,
    6,
    5,
    0,
    6,
    1,
    6
  ],
  [
    7,
    12,
    0,
    3,
    4,
    0,
    2,
    4,
    3,
    2,
    1,
    3,
    4,
    1,
    1,
    0,
    5,
    1,
    5,
    2,
    6,
    1,
    2,
    6,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    5,
    4,
    4,
    1,
    1,
    6,
    5,
    3,
    1,
    5,
    6,
    2
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    5,
    4,
    4,
    1,
    1,
    5,
    5,
    3,
    6,
    1,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    5,
    4,
    4,
    1,
    1,
    5,
    5,
    3,
    6,
    1,
    0,
    6
  ],
  [
    7,
    12,
    4,
    3,
    1,
    2,
    0,
    1,
    0,
    3,
    4,
    0,
    6,
    4,
    4,
    2,
    5,
    4,
    6,
    2,
    6,
    3,
    3,
    2,
    5,
    1
  ],
  [
    7,
    12,
    2,
    3,
    4,
    2,
    0,
    4,
    6,
    0,
    6,
    5,
    4,
    5,
    1,
    3,
    5,
    1,
    0,
    5,
    6,
    1,
    3,
    6,
    5,
    3
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    3,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    4,
    3,
    5,
    3,
    6,
    1,
    3,
    6,
    6,
    2,
    0,
    6,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    2,
    4,
    0,
    2,
    6,
    1,
    3,
    1,
    3,
    2,
    4,
    1,
    5,
    1,
    5,
    2,
    5,
    3,
    0,
    3,
    4,
    6
  ],
  [
    7,
    12,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    3,
    1,
    6
  ],
  [
    7,
    12,
    4,
    3,
    1,
    2,
    5,
    0,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    5,
    1,
    6,
    2,
    6,
    3,
    3,
    2,
    6,
    4
  ],
  [
    7,
    12,
    2,
    3,
    4,
    2,
    5,
    4,
    0,
    5,
    6,
    0,
    1,
    6,
    3,
    1,
    6,
    3,
    5,
    3,
    1,
    5,
    4,
    0,
    3,
    0
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    3,
    2,
    1,
    2,
    5,
    0,
    0,
    3,
    4,
    0,
    4,
    1,
    6,
    4,
    5,
    1,
    6,
    2,
    6,
    3,
    6,
    1,
    0,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    6,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    1,
    2,
    1,
    4,
    1,
    6,
    2,
    4,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    3,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    5,
    4,
    4,
    1,
    3,
    4,
    5,
    3,
    2,
    6,
    6,
    1
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    5,
    4,
    4,
    1,
    3,
    4,
    5,
    3,
    6,
    1,
    6,
    5
  ],
  [
    7,
    12,
    0,
    2,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    1,
    2,
    4,
    5,
    2,
    0,
    5,
    4,
    3,
    6,
    5,
    6,
    4,
    3,
    5
  ],
  [
    7,
    12,
    0,
    2,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    2,
    3,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    2,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    2,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    5,
    6
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    3,
    0,
    2,
    3,
    4,
    2,
    0,
    4,
    5,
    1,
    5,
    2,
    6,
    1,
    6,
    0,
    3,
    6,
    5,
    3,
    4,
    5,
    6,
    4
  ],
  [
    7,
    12,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    1,
    0,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    3,
    0,
    2,
    3,
    4,
    2,
    0,
    4,
    5,
    1,
    5,
    2,
    6,
    1,
    6,
    0,
    3,
    6,
    5,
    3,
    6,
    4,
    1,
    3
  ],
  [
    7,
    12,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    2,
    3,
    4,
    2,
    5,
    2,
    4,
    1,
    6,
    0,
    3,
    0,
    3,
    1,
    6,
    3,
    5,
    6,
    1,
    5,
    4,
    0,
    3,
    4
  ],
  [
    7,
    12,
    2,
    3,
    4,
    2,
    4,
    1,
    2,
    5,
    6,
    0,
    6,
    4,
    3,
    1,
    6,
    3,
    0,
    3,
    1,
    5,
    4,
    0,
    5,
    3
  ],
  [
    7,
    12,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    3,
    2,
    6,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    3,
    0,
    2,
    3,
    4,
    2,
    0,
    4,
    5,
    1,
    5,
    2,
    6,
    1,
    6,
    0,
    3,
    6,
    1,
    3,
    6,
    4,
    5,
    4
  ],
  [
    7,
    12,
    6,
    3,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    0,
    6,
    4,
    1,
    3,
    4,
    6,
    5,
    5,
    1,
    0,
    4
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    5,
    3,
    4,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    12,
    0,
    4,
    3,
    0,
    1,
    3,
    4,
    1,
    1,
    0,
    4,
    5,
    2,
    4,
    6,
    2,
    5,
    6,
    2,
    5,
    3,
    2,
    6,
    3
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    5,
    2,
    5,
    4,
    6,
    4,
    6,
    3,
    6,
    2,
    5,
    3
  ],
  [
    7,
    12,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    4,
    5,
    4,
    6
  ],
  [
    7,
    12,
    3,
    0,
    4,
    2,
    3,
    1,
    4,
    0,
    5,
    2,
    5,
    1,
    4,
    3,
    5,
    4,
    3,
    5,
    6,
    1,
    0,
    6,
    2,
    6
  ],
  [
    7,
    12,
    1,
    0,
    4,
    1,
    0,
    4,
    5,
    0,
    6,
    5,
    1,
    6,
    3,
    4,
    5,
    3,
    2,
    3,
    5,
    2,
    6,
    3,
    2,
    6
  ],
  [
    7,
    12,
    0,
    1,
    2,
    0,
    2,
    3,
    3,
    4,
    0,
    4,
    0,
    5,
    6,
    1,
    4,
    6,
    6,
    5,
    2,
    6,
    3,
    1,
    5,
    3
  ],
  [
    7,
    12,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    6,
    0,
    6,
    1,
    6,
    2,
    6,
    3,
    6,
    4,
    6,
    5
  ],
  [
    7,
    12,
    3,
    6,
    1,
    2,
    0,
    6,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    1,
    2,
    5,
    4,
    5,
    6,
    4
  ],
  [
    7,
    13,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    4,
    3,
    0,
    2,
    6,
    4
  ],
  [
    7,
    13,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    0,
    4,
    1,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    4,
    5,
    1,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    4,
    5,
    0,
    4,
    1,
    3,
    4,
    1,
    2,
    4,
    0,
    3,
    5,
    3,
    4,
    3,
    0,
    2,
    5,
    6
  ],
  [
    7,
    13,
    0,
    5,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    5,
    6,
    0,
    5,
    6,
    0,
    4,
    6,
    5,
    4,
    1,
    5,
    6,
    1,
    3,
    6,
    5,
    3,
    2,
    5,
    6,
    2,
    1,
    0,
    2,
    1
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    6
  ],
  [
    7,
    13,
    3,
    4,
    0,
    3,
    4,
    0,
    1,
    4,
    3,
    1,
    2,
    3,
    4,
    2,
    1,
    0,
    2,
    1,
    6,
    0,
    3,
    6,
    5,
    3,
    4,
    5
  ],
  [
    7,
    13,
    3,
    4,
    0,
    3,
    4,
    0,
    1,
    4,
    3,
    1,
    2,
    3,
    4,
    2,
    1,
    0,
    2,
    1,
    6,
    0,
    3,
    6,
    5,
    3,
    0,
    5
  ],
  [
    7,
    13,
    3,
    4,
    0,
    3,
    4,
    0,
    1,
    4,
    3,
    1,
    2,
    3,
    4,
    2,
    1,
    0,
    2,
    1,
    6,
    4,
    0,
    6,
    5,
    0,
    3,
    5
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    5,
    1,
    6,
    5,
    1,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    6,
    1,
    3,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    5,
    3,
    6,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    1,
    6,
    5,
    5,
    1
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    4,
    0,
    5,
    6,
    0,
    3,
    6,
    6,
    4
  ],
  [
    7,
    13,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    1,
    5,
    1,
    2,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    5,
    6
  ],
  [
    7,
    13,
    1,
    0,
    2,
    1,
    4,
    2,
    1,
    4,
    3,
    1,
    0,
    3,
    2,
    3,
    5,
    2,
    1,
    5,
    0,
    5,
    6,
    0,
    1,
    6,
    2,
    6
  ],
  [
    7,
    13,
    2,
    5,
    6,
    2,
    5,
    6,
    4,
    5,
    3,
    4,
    0,
    3,
    4,
    0,
    1,
    4,
    3,
    1,
    6,
    3,
    2,
    1,
    4,
    2,
    3,
    2
  ],
  [
    7,
    13,
    0,
    3,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    2,
    4,
    3,
    2,
    1,
    3,
    4,
    1,
    0,
    4,
    3,
    0,
    6,
    3,
    1,
    6,
    5,
    1,
    4,
    5,
    6,
    4,
    3,
    5,
    1,
    2
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    4,
    2,
    5,
    3,
    5,
    3,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    5,
    3,
    6,
    4,
    5
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5
  ],
  [
    7,
    13,
    0,
    2,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    2,
    5,
    1,
    2,
    5,
    1,
    4,
    5,
    3,
    4,
    0,
    3,
    4,
    0,
    3,
    2,
    4,
    2,
    1,
    3,
    6,
    3,
    2,
    6,
    1,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    0,
    4,
    6,
    1,
    4,
    6
  ],
  [
    7,
    13,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    1,
    0,
    5,
    2,
    5,
    0,
    6,
    5,
    1,
    6,
    4,
    0
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    1,
    3,
    3,
    2,
    4,
    0,
    2,
    5,
    3,
    4,
    5,
    3,
    0,
    5,
    6,
    4,
    6,
    1
  ],
  [
    7,
    13,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    1,
    0,
    5,
    2,
    6,
    2,
    5,
    6,
    0,
    5,
    1,
    2
  ],
  [
    7,
    13,
    5,
    4,
    6,
    2,
    6,
    4,
    4,
    3,
    5,
    0,
    3,
    1,
    3,
    2,
    6,
    3,
    5,
    6,
    4,
    0,
    1,
    4,
    5,
    1,
    0,
    3
  ],
  [
    7,
    13,
    0,
    2,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    1,
    5,
    4,
    1,
    0,
    4,
    5,
    0,
    2,
    5,
    4,
    2,
    3,
    4,
    5,
    3,
    0,
    1,
    2,
    0,
    3,
    2,
    6,
    5,
    6,
    4
  ],
  [
    7,
    13,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    0,
    4,
    0,
    6,
    4,
    6
  ],
  [
    7,
    13,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    5,
    0,
    4,
    5,
    1,
    5,
    6,
    1,
    0,
    6,
    3,
    6
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    4,
    1,
    5,
    3,
    2,
    5,
    1,
    0,
    6,
    3,
    4,
    6,
    5,
    1
  ],
  [
    7,
    13,
    5,
    2,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    1,
    0,
    6,
    2,
    6,
    3,
    1,
    2,
    3,
    1,
    4,
    0
  ],
  [
    7,
    13,
    1,
    0,
    2,
    1,
    0,
    2,
    0,
    3,
    3,
    2,
    6,
    2,
    1,
    6,
    5,
    1,
    6,
    5,
    4,
    6,
    5,
    4,
    0,
    5,
    4,
    0
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    4,
    1,
    3,
    1,
    0,
    4,
    2,
    3,
    0,
    3,
    2,
    0,
    5,
    0,
    6,
    5,
    4,
    6,
    3,
    5,
    4,
    2
  ],
  [
    7,
    13,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    0,
    4,
    2,
    3,
    6,
    5,
    0,
    6
  ],
  [
    7,
    13,
    5,
    2,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    1,
    0,
    6,
    2,
    6,
    3,
    2,
    3,
    5,
    0,
    4,
    0
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    5,
    4,
    0,
    4,
    5,
    0,
    5,
    1,
    5,
    2,
    5,
    3,
    6,
    3,
    6,
    4,
    4,
    2,
    0,
    3
  ],
  [
    7,
    13,
    0,
    1,
    0,
    6,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    6,
    2,
    0,
    6,
    5,
    0,
    2,
    5,
    5,
    3,
    4,
    5,
    6,
    4,
    3,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    6,
    5,
    0,
    6,
    0,
    2,
    2,
    5,
    5,
    3,
    4,
    5,
    6,
    4,
    3,
    6
  ],
  [
    7,
    13,
    3,
    4,
    0,
    3,
    4,
    0,
    3,
    6,
    3,
    1,
    2,
    3,
    4,
    2,
    1,
    0,
    2,
    1,
    5,
    2,
    3,
    5,
    6,
    2,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    5,
    0,
    5,
    1,
    5,
    2,
    5,
    3,
    5,
    4,
    6,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    13,
    1,
    0,
    2,
    1,
    6,
    0,
    1,
    4,
    3,
    1,
    0,
    3,
    2,
    3,
    1,
    6,
    1,
    5,
    2,
    6,
    4,
    3,
    5,
    4,
    6,
    5
  ],
  [
    7,
    13,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    6,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    2,
    5,
    0,
    6,
    3,
    4,
    1,
    3,
    4,
    6,
    5,
    1,
    5,
    3,
    5,
    1,
    3
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    6,
    4,
    4,
    3,
    5,
    4,
    5,
    2,
    6,
    0,
    3,
    6,
    3,
    5
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    6,
    0,
    4,
    6,
    5,
    0,
    6,
    6,
    4,
    2,
    5,
    5,
    3,
    4,
    5,
    1,
    5,
    6,
    1
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    2,
    3,
    5,
    2,
    6,
    5,
    3,
    6,
    4,
    3,
    5,
    4,
    0,
    5,
    2,
    0,
    1,
    2,
    0,
    1,
    5,
    1,
    4,
    2,
    6,
    4
  ],
  [
    7,
    13,
    2,
    1,
    0,
    5,
    6,
    0,
    4,
    6,
    5,
    4,
    1,
    5,
    6,
    1,
    3,
    6,
    5,
    3,
    2,
    5,
    6,
    2,
    1,
    0,
    3,
    4
  ],
  [
    7,
    13,
    0,
    1,
    2,
    0,
    2,
    3,
    3,
    4,
    0,
    4,
    0,
    5,
    6,
    1,
    4,
    6,
    6,
    5,
    2,
    6,
    3,
    1,
    5,
    3,
    6,
    0
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    4,
    6,
    5,
    1,
    4,
    5,
    3,
    1,
    5,
    2,
    6,
    0,
    6,
    1,
    2,
    1,
    4,
    1
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    3,
    5,
    4,
    5,
    2,
    5,
    1,
    6,
    5,
    1,
    6,
    2,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    13,
    3,
    0,
    2,
    3,
    4,
    2,
    0,
    4,
    5,
    1,
    5,
    2,
    6,
    1,
    6,
    0,
    3,
    6,
    1,
    3,
    6,
    4,
    5,
    4,
    5,
    3
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    0,
    2,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    6,
    2,
    3,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    3,
    4,
    5,
    0,
    3,
    5,
    0,
    6,
    6,
    4,
    2,
    3,
    4,
    2,
    1,
    0,
    2,
    1,
    1,
    6,
    5,
    1,
    2,
    5,
    6,
    2
  ],
  [
    7,
    13,
    0,
    2,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    13,
    3,
    0,
    2,
    3,
    4,
    2,
    0,
    4,
    5,
    3,
    5,
    2,
    5,
    4,
    1,
    3,
    1,
    0,
    4,
    1,
    6,
    0,
    3,
    6,
    1,
    6
  ],
  [
    7,
    13,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    6,
    0,
    1,
    6,
    6,
    3,
    4,
    6,
    1,
    2,
    5,
    0
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5
  ],
  [
    7,
    13,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    5,
    3,
    5,
    4,
    6,
    4,
    6,
    2,
    6,
    5,
    5,
    2,
    3,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    13,
    0,
    1,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    13,
    0,
    1,
    2,
    0,
    2,
    3,
    3,
    4,
    0,
    4,
    0,
    5,
    6,
    1,
    4,
    6,
    6,
    5,
    2,
    6,
    3,
    1,
    5,
    3,
    2,
    1
  ],
  [
    7,
    14,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    1,
    3,
    2,
    0,
    4,
    0,
    5,
    3
  ],
  [
    7,
    14,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    2,
    3,
    0,
    4,
    1,
    6
  ],
  [
    7,
    14,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    1,
    2,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    6,
    5,
    3,
    4,
    5,
    6,
    4
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    6,
    2,
    1,
    6,
    5,
    1,
    0,
    5
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    5,
    2,
    3,
    5,
    6,
    4,
    0,
    6
  ],
  [
    7,
    14,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    2,
    4,
    5,
    1,
    0,
    3,
    1,
    4,
    0,
    1,
    0,
    4,
    6,
    4,
    0,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    3,
    4,
    3,
    5,
    4,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    6,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    3,
    4,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    0,
    1,
    5,
    1,
    0,
    4,
    1,
    4,
    5,
    3,
    2,
    5,
    6,
    4,
    0,
    6
  ],
  [
    7,
    14,
    1,
    3,
    2,
    1,
    0,
    2,
    5,
    0,
    4,
    5,
    3,
    4,
    5,
    3,
    2,
    5,
    1,
    0,
    4,
    1,
    6,
    1,
    5,
    1,
    2,
    6,
    0,
    4
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    6,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    6,
    0,
    3,
    4,
    3,
    1,
    5,
    4,
    5,
    0,
    0,
    3,
    1,
    5,
    5,
    3,
    6,
    4,
    6,
    1
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    5,
    3,
    6,
    5,
    4,
    6,
    5,
    4
  ],
  [
    7,
    14,
    3,
    1,
    1,
    4,
    2,
    3,
    3,
    4,
    0,
    4,
    1,
    5,
    0,
    1,
    0,
    2,
    2,
    5,
    5,
    3,
    4,
    5,
    1,
    2,
    6,
    2,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    6,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    4,
    6,
    5,
    2,
    1,
    5,
    0,
    5,
    6,
    3
  ],
  [
    7,
    14,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    5,
    0,
    5,
    1,
    5,
    2,
    5,
    3,
    5,
    4,
    4,
    2,
    3,
    0,
    6,
    1,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    1,
    2,
    5,
    2,
    4,
    0,
    3,
    1,
    5,
    0,
    6,
    5,
    6,
    4,
    0,
    1
  ],
  [
    7,
    14,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    5,
    2,
    6,
    0,
    6,
    1,
    5,
    0,
    1,
    2,
    3,
    1,
    4,
    0
  ],
  [
    7,
    14,
    5,
    6,
    0,
    5,
    6,
    0,
    4,
    6,
    5,
    4,
    1,
    5,
    6,
    1,
    3,
    6,
    5,
    3,
    2,
    5,
    6,
    2,
    1,
    0,
    2,
    1,
    3,
    4
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    2,
    0,
    2,
    3,
    3,
    4,
    0,
    4,
    0,
    5,
    6,
    1,
    4,
    6,
    6,
    5,
    2,
    6,
    3,
    1,
    5,
    3,
    6,
    0,
    3,
    6
  ],
  [
    7,
    14,
    3,
    1,
    4,
    2,
    4,
    5,
    4,
    0,
    1,
    4,
    0,
    3,
    5,
    0,
    5,
    2,
    6,
    1,
    3,
    6,
    6,
    0,
    5,
    6,
    6,
    2,
    4,
    6
  ],
  [
    7,
    14,
    0,
    4,
    3,
    0,
    2,
    3,
    4,
    2,
    1,
    4,
    3,
    1,
    1,
    0,
    2,
    1,
    5,
    4,
    1,
    5,
    6,
    1,
    3,
    6,
    0,
    5,
    6,
    0
  ],
  [
    7,
    14,
    3,
    4,
    4,
    2,
    1,
    5,
    4,
    0,
    1,
    4,
    5,
    3,
    3,
    0,
    5,
    2,
    6,
    4,
    0,
    6,
    3,
    6,
    2,
    6,
    5,
    6,
    1,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    6,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    4,
    5,
    3,
    4,
    3,
    1,
    5,
    2,
    1,
    6,
    5,
    6,
    6,
    0,
    5,
    3,
    6,
    4,
    0,
    1
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    1,
    4,
    6,
    0,
    3,
    1,
    5,
    2,
    4,
    5,
    5,
    6,
    1,
    5,
    5,
    3,
    6,
    4,
    3,
    0
  ],
  [
    7,
    14,
    3,
    1,
    4,
    2,
    0,
    3,
    4,
    0,
    1,
    4,
    5,
    3,
    5,
    0,
    5,
    2,
    6,
    4,
    1,
    6,
    6,
    3,
    0,
    6,
    6,
    5,
    2,
    6
  ],
  [
    7,
    14,
    0,
    1,
    4,
    2,
    3,
    0,
    4,
    0,
    4,
    5,
    5,
    3,
    1,
    3,
    5,
    2,
    6,
    4,
    2,
    6,
    6,
    5,
    3,
    6,
    6,
    1,
    0,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    6,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    1,
    4,
    6,
    1,
    3,
    1,
    5,
    2,
    5,
    0,
    5,
    6,
    4,
    5,
    5,
    3,
    6,
    0,
    3,
    0
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    3,
    0,
    4,
    0,
    4,
    5,
    3,
    4,
    3,
    1,
    5,
    2,
    0,
    1,
    5,
    6,
    6,
    0,
    5,
    3,
    6,
    4,
    1,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    0,
    4,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    0,
    4,
    0,
    6,
    1,
    3,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    1,
    4,
    4,
    5,
    3,
    1,
    5,
    2,
    5,
    0,
    6,
    1,
    0,
    6,
    5,
    3,
    6,
    4,
    3,
    0
  ],
  [
    7,
    14,
    0,
    1,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    5,
    6
  ],
  [
    7,
    14,
    2,
    3,
    4,
    2,
    6,
    3,
    4,
    0,
    4,
    5,
    3,
    5,
    3,
    1,
    5,
    2,
    3,
    0,
    1,
    4,
    6,
    0,
    1,
    6,
    6,
    4,
    0,
    1
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    14,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    6,
    4,
    6,
    5,
    6
  ],
  [
    7,
    14,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    5,
    6,
    0,
    6,
    0,
    2,
    5,
    0,
    3,
    5,
    1,
    3,
    6,
    1,
    4,
    6,
    2,
    4
  ],
  [
    7,
    14,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    6,
    4,
    5
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    4,
    5
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    1,
    3,
    2,
    0,
    4,
    0,
    5,
    3,
    1,
    6
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    2,
    4,
    5,
    2,
    1,
    5,
    1,
    4,
    1,
    3,
    2,
    0,
    4,
    0,
    5,
    3,
    0,
    6
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    5,
    6
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    3,
    6,
    4,
    6,
    5,
    6,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    1,
    6,
    0,
    1
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    6,
    1,
    4,
    6
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    1,
    4,
    2,
    4,
    3,
    5,
    1,
    0,
    5,
    5,
    2,
    3,
    5,
    4,
    5,
    6,
    1,
    5,
    6
  ],
  [
    7,
    15,
    4,
    3,
    4,
    5,
    5,
    3,
    0,
    1,
    0,
    5,
    0,
    3,
    2,
    4,
    1,
    5,
    1,
    3,
    6,
    5,
    3,
    6,
    6,
    0,
    1,
    6,
    6,
    2,
    4,
    6
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    5,
    6,
    0,
    4,
    0,
    5,
    0,
    6,
    3,
    6,
    1,
    3,
    4,
    6,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    1,
    6,
    0,
    1
  ],
  [
    7,
    15,
    0,
    2,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    6,
    1,
    4,
    5,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    4,
    6
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    0,
    3,
    0,
    4,
    0,
    5,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    0,
    3,
    0,
    4,
    6,
    0,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    0,
    2,
    3,
    0,
    1,
    3,
    2,
    3,
    5,
    1,
    3,
    5,
    4,
    3,
    2,
    4,
    6,
    2,
    3,
    6,
    6,
    1,
    0,
    5,
    4,
    0
  ],
  [
    7,
    15,
    0,
    1,
    2,
    0,
    3,
    2,
    4,
    3,
    1,
    4,
    3,
    1,
    4,
    2,
    0,
    4,
    3,
    0,
    6,
    3,
    4,
    6,
    5,
    4,
    3,
    5,
    6,
    2,
    5,
    1
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    1,
    2,
    1,
    3,
    1,
    4,
    2,
    3,
    2,
    4,
    3,
    4,
    5,
    3,
    4,
    5,
    6,
    4,
    5,
    6,
    6,
    3
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    0,
    3,
    4,
    0,
    4,
    3,
    4,
    2,
    6,
    1,
    2,
    6,
    5,
    2,
    4,
    5,
    6,
    4,
    0,
    6,
    6,
    5,
    3,
    6
  ],
  [
    7,
    15,
    0,
    1,
    5,
    3,
    1,
    3,
    0,
    4,
    3,
    0,
    4,
    3,
    2,
    4,
    5,
    2,
    4,
    5,
    6,
    4,
    2,
    6,
    6,
    5,
    3,
    6,
    6,
    1,
    0,
    6
  ],
  [
    7,
    15,
    5,
    2,
    4,
    5,
    3,
    1,
    0,
    4,
    0,
    5,
    0,
    3,
    2,
    4,
    1,
    5,
    1,
    4,
    6,
    3,
    1,
    6,
    6,
    0,
    5,
    6,
    6,
    2,
    4,
    6
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    4,
    5,
    0,
    5,
    0,
    3,
    2,
    0,
    3,
    1,
    6,
    4,
    5,
    6,
    6,
    3,
    0,
    6,
    6,
    2,
    1,
    6
  ],
  [
    7,
    15,
    5,
    2,
    3,
    0,
    5,
    3,
    0,
    4,
    0,
    5,
    4,
    3,
    2,
    4,
    1,
    5,
    1,
    4,
    6,
    4,
    2,
    6,
    6,
    0,
    5,
    6,
    6,
    3,
    1,
    6
  ],
  [
    7,
    15,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    6,
    1,
    4,
    5,
    0,
    3,
    0,
    4,
    0,
    5,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    0,
    6,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    3,
    4,
    0,
    1,
    0,
    3,
    0,
    4,
    0,
    5,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    1,
    6,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    6
  ],
  [
    7,
    15,
    5,
    2,
    4,
    5,
    5,
    3,
    0,
    4,
    0,
    1,
    1,
    3,
    2,
    4,
    3,
    0,
    1,
    4,
    6,
    4,
    1,
    6,
    6,
    0,
    3,
    6,
    6,
    2,
    5,
    6
  ],
  [
    7,
    15,
    5,
    0,
    4,
    3,
    5,
    3,
    5,
    2,
    0,
    1,
    1,
    3,
    2,
    4,
    3,
    0,
    1,
    4,
    6,
    2,
    5,
    6,
    6,
    4,
    3,
    6,
    6,
    0,
    1,
    6
  ],
  [
    7,
    15,
    3,
    4,
    4,
    5,
    0,
    3,
    4,
    6,
    0,
    1,
    1,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    0,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    0,
    2,
    0,
    3,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    15,
    3,
    4,
    5,
    0,
    0,
    3,
    0,
    4,
    4,
    6,
    1,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    1,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    5,
    2
  ],
  [
    7,
    15,
    6,
    4,
    5,
    2,
    0,
    3,
    0,
    4,
    2,
    4,
    1,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    3,
    5,
    2,
    3,
    0,
    1,
    5,
    6,
    4,
    5
  ],
  [
    7,
    15,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    0,
    1,
    0,
    2,
    0,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    15,
    2,
    3,
    0,
    2,
    3,
    0,
    4,
    3,
    1,
    4,
    5,
    1,
    4,
    5,
    1,
    0,
    5,
    2,
    6,
    2,
    5,
    6,
    6,
    1,
    0,
    6,
    6,
    4,
    3,
    6
  ],
  [
    7,
    15,
    3,
    0,
    3,
    5,
    3,
    4,
    2,
    0,
    2,
    5,
    2,
    4,
    1,
    4,
    1,
    5,
    1,
    0,
    6,
    0,
    1,
    6,
    6,
    5,
    3,
    6,
    6,
    4,
    2,
    6
  ],
  [
    7,
    15,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    15,
    3,
    4,
    6,
    2,
    0,
    3,
    0,
    4,
    5,
    0,
    1,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    4,
    5,
    2,
    3,
    2,
    4,
    5,
    1,
    5,
    2
  ],
  [
    7,
    15,
    3,
    4,
    6,
    2,
    0,
    3,
    0,
    4,
    5,
    0,
    5,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    0,
    1,
    4,
    6,
    2,
    3,
    2,
    4,
    5,
    1,
    5,
    2
  ],
  [
    7,
    15,
    0,
    1,
    1,
    2,
    2,
    3,
    3,
    4,
    0,
    4,
    6,
    2,
    1,
    6,
    6,
    0,
    4,
    6,
    5,
    4,
    0,
    5,
    3,
    5,
    6,
    3,
    5,
    2,
    1,
    5
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    4,
    5,
    2,
    6
  ],
  [
    7,
    16,
    3,
    0,
    4,
    1,
    4,
    3,
    1,
    3,
    4,
    0,
    2,
    5,
    6,
    2,
    5,
    6,
    1,
    5,
    4,
    5,
    3,
    5,
    0,
    5,
    0,
    6,
    3,
    6,
    4,
    6,
    6,
    1
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    16,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    16,
    3,
    4,
    5,
    1,
    0,
    3,
    0,
    4,
    5,
    0,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    4,
    5,
    2,
    5
  ],
  [
    7,
    16,
    2,
    4,
    3,
    1,
    3,
    0,
    4,
    3,
    4,
    0,
    5,
    2,
    4,
    5,
    5,
    0,
    3,
    5,
    5,
    1,
    6,
    5,
    1,
    6,
    3,
    6,
    6,
    0,
    4,
    6,
    6,
    2
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    16,
    2,
    4,
    4,
    1,
    3,
    0,
    3,
    1,
    4,
    0,
    5,
    2,
    4,
    5,
    5,
    0,
    3,
    5,
    6,
    5,
    1,
    5,
    6,
    1,
    3,
    6,
    4,
    6,
    6,
    2,
    6,
    0
  ],
  [
    7,
    16,
    0,
    1,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    16,
    2,
    5,
    0,
    1,
    4,
    5,
    1,
    3,
    5,
    0,
    4,
    3,
    5,
    3,
    2,
    4,
    1,
    4,
    3,
    0,
    6,
    3,
    2,
    6,
    6,
    4,
    5,
    6,
    6,
    1,
    0,
    6
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    3,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    16,
    2,
    5,
    5,
    1,
    3,
    1,
    0,
    4,
    5,
    0,
    4,
    3,
    5,
    3,
    2,
    4,
    1,
    4,
    3,
    0,
    6,
    2,
    4,
    6,
    5,
    6,
    6,
    1,
    0,
    6,
    3,
    6
  ],
  [
    7,
    16,
    1,
    6,
    0,
    1,
    0,
    3,
    0,
    4,
    5,
    0,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    3,
    5,
    2,
    3,
    2,
    4,
    5,
    6,
    4,
    5,
    2,
    5
  ],
  [
    7,
    16,
    3,
    4,
    5,
    1,
    0,
    3,
    0,
    4,
    5,
    0,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    1,
    6,
    2,
    3,
    2,
    4,
    5,
    6,
    0,
    1,
    2,
    5
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    5,
    6
  ],
  [
    7,
    16,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    5,
    6
  ],
  [
    7,
    16,
    2,
    5,
    5,
    1,
    3,
    5,
    0,
    4,
    0,
    1,
    4,
    3,
    3,
    2,
    2,
    4,
    1,
    4,
    0,
    5,
    6,
    4,
    2,
    6,
    6,
    3,
    5,
    6,
    6,
    1,
    0,
    6
  ],
  [
    7,
    16,
    5,
    6,
    5,
    1,
    0,
    3,
    0,
    4,
    0,
    1,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    3,
    5,
    2,
    3,
    2,
    4,
    6,
    2,
    4,
    5,
    2,
    5
  ],
  [
    7,
    16,
    3,
    4,
    5,
    1,
    0,
    3,
    0,
    4,
    0,
    1,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    5,
    0,
    2,
    3,
    2,
    4,
    6,
    2,
    6,
    5,
    2,
    5
  ],
  [
    7,
    16,
    5,
    0,
    5,
    1,
    0,
    3,
    0,
    4,
    6,
    1,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    6,
    0,
    3,
    5,
    2,
    3,
    2,
    4,
    6,
    2,
    4,
    5,
    2,
    5
  ],
  [
    7,
    17,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    4,
    5,
    6,
    2,
    1,
    6
  ],
  [
    7,
    17,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    17,
    4,
    0,
    4,
    3,
    0,
    1,
    3,
    0,
    2,
    4,
    3,
    1,
    5,
    3,
    4,
    5,
    5,
    2,
    6,
    5,
    5,
    0,
    1,
    5,
    6,
    1,
    0,
    6,
    6,
    4,
    2,
    6,
    3,
    6
  ],
  [
    7,
    17,
    0,
    1,
    5,
    1,
    5,
    3,
    0,
    4,
    5,
    0,
    4,
    3,
    3,
    1,
    2,
    5,
    1,
    4,
    3,
    0,
    2,
    4,
    6,
    2,
    5,
    6,
    6,
    3,
    1,
    6,
    6,
    0,
    4,
    6
  ],
  [
    7,
    17,
    3,
    4,
    5,
    1,
    0,
    3,
    0,
    4,
    4,
    5,
    4,
    6,
    3,
    6,
    1,
    3,
    1,
    4,
    0,
    1,
    3,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    5,
    0,
    5,
    6,
    6,
    2
  ],
  [
    7,
    17,
    3,
    2,
    4,
    1,
    0,
    1,
    3,
    0,
    2,
    4,
    4,
    3,
    5,
    1,
    4,
    5,
    5,
    2,
    0,
    5,
    5,
    3,
    6,
    5,
    2,
    6,
    6,
    3,
    0,
    6,
    1,
    6,
    4,
    6
  ],
  [
    7,
    17,
    3,
    2,
    4,
    1,
    4,
    0,
    3,
    0,
    2,
    4,
    3,
    1,
    5,
    2,
    4,
    5,
    5,
    0,
    3,
    5,
    5,
    1,
    6,
    5,
    2,
    6,
    6,
    0,
    3,
    6,
    6,
    4,
    1,
    6
  ],
  [
    7,
    17,
    3,
    2,
    5,
    1,
    5,
    0,
    0,
    4,
    0,
    1,
    4,
    3,
    5,
    3,
    2,
    5,
    1,
    4,
    3,
    0,
    2,
    4,
    6,
    4,
    5,
    6,
    6,
    2,
    3,
    6,
    6,
    0,
    1,
    6
  ],
  [
    7,
    17,
    3,
    2,
    5,
    1,
    5,
    0,
    0,
    4,
    4,
    5,
    0,
    1,
    3,
    1,
    2,
    5,
    1,
    4,
    3,
    0,
    2,
    4,
    6,
    0,
    3,
    6,
    6,
    1,
    5,
    6,
    6,
    2,
    4,
    6
  ],
  [
    7,
    17,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6
  ],
  [
    7,
    18,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    3,
    4,
    3,
    5,
    4,
    5,
    6,
    1,
    0,
    6,
    5,
    6
  ],
  [
    7,
    18,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6
  ],
  [
    7,
    18,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    4,
    1,
    5,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    18,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    18,
    4,
    0,
    4,
    5,
    3,
    0,
    3,
    5,
    2,
    0,
    2,
    5,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    0,
    2,
    3,
    2,
    4,
    6,
    0,
    5,
    6,
    6,
    1,
    2,
    6,
    6,
    4,
    3,
    6
  ],
  [
    7,
    19,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    19,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    20,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
  [
    7,
    21,
    0,
    1,
    0,
    2,
    0,
    3,
    0,
    4,
    0,
    5,
    0,
    6,
    1,
    2,
    1,
    3,
    1,
    4,
    1,
    5,
    1,
    6,
    2,
    3,
    2,
    4,
    2,
    5,
    2,
    6,
    3,
    4,
    3,
    5,
    3,
    6,
    4,
    5,
    4,
    6,
    5,
    6
  ],
];
